import pytest
import construct
from xbox.sg.utils import adapters


@pytest.mark.skip(reason="Not Implemented")
def test_cryptotunnel():
    pass


def test_json():
    import json

    test_data = {"Z": "ABC", "A": "XYZ", "B": 23}
    bytes_data = json.dumps(test_data,
                            separators=(',', ':'),
                            sort_keys=True).encode('utf-8')
    adapter = adapters.JsonAdapter(construct.GreedyString("utf8"))

    parsed = adapter.parse(bytes_data)
    built = adapter.build(test_data)

    with pytest.raises(json.JSONDecodeError):
        adapter.parse(b'invalid data')

    with pytest.raises(TypeError):
        adapter.parse(234)

    with pytest.raises(TypeError):
        adapter.parse("string")

    with pytest.raises(TypeError):
        adapter.build(b'invalid data')

    with pytest.raises(TypeError):
        adapter.build(234)

    with pytest.raises(TypeError):
        adapter.build("string")

    assert parsed == test_data
    assert built == bytes_data


def test_uuid(uuid_dummy):
    import struct
    test_data = uuid_dummy
    uuid_string = str(test_data).upper()
    uuid_stringbytes = uuid_string.encode('utf-8')
    uuid_sgstring = struct.pack('>H', len(uuid_stringbytes)) + uuid_stringbytes + b'\x00'
    uuid_bytes = test_data.bytes

    adapter_bytes = adapters.UUIDAdapter()
    parsed_bytes = adapter_bytes.parse(uuid_bytes)
    built_bytes = adapter_bytes.build(test_data)

    adapter_string = adapters.UUIDAdapter("utf-8")
    parsed_sgstring = adapter_string.parse(uuid_sgstring)
    built_sgstring = adapter_string.build(test_data)

    adapter_invalid = adapters.UUIDAdapter("utf-invalid")

    with pytest.raises(LookupError):
        adapter_invalid.parse(uuid_sgstring)

    with pytest.raises(LookupError):
        adapter_invalid.build(test_data)

    with pytest.raises(construct.StreamError):
        adapter_bytes.parse(uuid_bytes[:-2])

    with pytest.raises(TypeError):
        adapter_bytes.parse('string, not bytes object')

    with pytest.raises(TypeError):
        adapter_bytes.build('some-string, not UUID object')

    with pytest.raises(construct.StreamError):
        adapter_string.parse(uuid_sgstring[:-3])

    with pytest.raises(TypeError):
        adapter_string.parse('string, not sgstring-bytes')

    with pytest.raises(TypeError):
        adapter_string.build('some-string, not UUID object')

    assert parsed_bytes == test_data
    assert built_bytes == uuid_bytes
    assert parsed_sgstring == test_data
    assert built_sgstring == uuid_sgstring


def test_certificate(certificate_data):
    import struct
    prefixed_data = struct.pack('>H', len(certificate_data)) + certificate_data
    certinfo = adapters.CertificateInfo(certificate_data)

    adapter = adapters.CertificateAdapter()

    parsed = adapter.parse(prefixed_data)
    built = adapter.build(certinfo)

    with pytest.raises(construct.core.StreamError):
        adapter.parse(prefixed_data[:-6])

    with pytest.raises(construct.core.StreamError):
        adapter.parse(b'\xAB' * 10)

    with pytest.raises(construct.core.StreamError):
        adapter.parse(certificate_data)

    with pytest.raises(TypeError):
        adapter.parse('string, not bytes')

    with pytest.raises(TypeError):
        adapter.parse(123)

    with pytest.raises(TypeError):
        adapter.build(b'\xAB' * 10)

    with pytest.raises(TypeError):
        adapter.build(certificate_data)

    with pytest.raises(TypeError):
        adapter.build('string, not bytes')

    with pytest.raises(TypeError):
        adapter.build(123)

    assert isinstance(parsed, adapters.CertificateInfo) is True
    assert parsed == certinfo
    assert built == prefixed_data


def test_certificateinfo(certificate_data):
    certinfo = adapters.CertificateInfo(certificate_data)

    with pytest.raises(ValueError):
        adapters.CertificateInfo(certificate_data[:-1])

    with pytest.raises(ValueError):
        adapters.CertificateInfo(b'\x23' * 10)

    with pytest.raises(TypeError):
        adapters.CertificateInfo('some string')

    with pytest.raises(TypeError):
        adapters.CertificateInfo(123)

    with pytest.raises(TypeError):
        certinfo.dump(encoding='invalid param')

    assert certinfo.dump() == certificate_data
    assert certinfo.liveid == 'FFFFFFFFFFF'



import io

from xbox.sg.crypto import PKCS7Padding
from xbox.auxiliary import packer
from xbox.auxiliary import packet


def _read_aux_packets(data):
    with io.BytesIO(data) as stream:
        while stream.tell() < len(data):
            header_data = stream.read(4)
            header = packet.aux_header_struct.parse(header_data)

            if header.magic != packet.AUX_PACKET_MAGIC:
                raise Exception('Invalid packet magic received from console')

            padded_payload_sz = header.payload_size + PKCS7Padding.size(header.payload_size, 16)

            payload_data = stream.read(padded_payload_sz)
            hmac = stream.read(32)
            yield header_data + payload_data + hmac


def test_client_unpack(aux_streams, aux_crypto):
    data = aux_streams['fo4_client_to_console']
    for msg in _read_aux_packets(data):
        packer.unpack(msg, aux_crypto, client_data=True)


def test_server_unpack(aux_streams, aux_crypto):
    data = aux_streams['fo4_console_to_client']
    for msg in _read_aux_packets(data):
        packer.unpack(msg, aux_crypto)


def test_decryption(aux_streams, aux_crypto):
    data = aux_streams['fo4_console_to_client']
    messages = list(_read_aux_packets(data))
    # Need to unpack messages in order, starting with the first one
    # -> Gets IV from previous decryption
    packer.unpack(messages[0], aux_crypto)
    json_msg = packer.unpack(messages[1], aux_crypto)
    assert json_msg == b'{"lang":"de","version":"1.10.52.0"}\n'



from http import HTTPStatus


def test_device_overview(rest_client):
    resp = rest_client.test_client().get('/device')

    assert resp.status_code == HTTPStatus.OK
    assert resp.json['success'] is True


def test_device_poweron(rest_client, console_liveid):
    resp = rest_client.test_client().get('/device/{0}/poweron'.format(console_liveid))

    assert resp.status_code == HTTPStatus.OK
    assert resp.json['success'] is True


def test_device_info():
    pass


def test_connect():
    pass


def test_disconnect():
    pass


def test_poweroff():
    pass


def test_launch_title():
    pass


def test_infrared():
    pass


def test_infrared_available_keys():
    pass


def test_infrared_send():
    pass


def test_media_overview():
    pass


def test_media_command():
    pass


def test_media_command_seek():
    pass


def test_input_overview():
    pass


def test_input_send_button():
    pass


def test_stump_headend_info():
    pass


def test_stump_livetv_info():
    pass


def test_stump_tuner_lineups():
    pass


def test_text_overview():
    pass


def test_text_send():
    pass


def test_media_status(rest_client_connected_media_console_status, console_liveid):
    client = rest_client_connected_media_console_status
    resp = client.test_client().get('/device/{0}/media_status'.format(console_liveid))

    assert resp.status_code == HTTPStatus.OK
    assert resp.json['success'] is True
    assert resp.json['media_status'] is not None
    media_status = resp.json['media_status']

    assert media_status['title_id'] == 274278798
    assert media_status['aum_id'] == 'AIVDE_s9eep9cpjhg6g!App'
    assert media_status['asset_id'] == ''
    assert media_status['media_type'] == 'Video'
    assert media_status['sound_level'] == 'Full'
    assert media_status['enabled_commands'] == 6
    assert media_status['playback_status'] == 'Playing'
    assert media_status['rate'] == 1.00
    assert media_status['position'] == 0
    assert media_status['media_start'] == 0
    assert media_status['media_end'] == 0
    assert media_status['min_seek'] == 0
    assert media_status['max_seek'] == 0
    assert media_status['metadata'] is not None
    metadata = media_status['metadata']

    assert len(metadata) == 2
    assert 'title' in metadata
    assert 'subtitle' in metadata
    assert metadata['title'] == 'Some Movietitle'
    assert metadata['subtitle'] == ''


def test_console_status(rest_client_connected_media_console_status, console_liveid):
    client = rest_client_connected_media_console_status
    resp = client.test_client().get('/device/{0}/console_status'.format(console_liveid))

    assert resp.status_code == HTTPStatus.OK
    assert resp.json['success'] is True
    assert resp.json['console_status'] is not None
    status = resp.json['console_status']

    assert status['kernel_version'] == '10.0.14393'
    assert status['live_tv_provider'] == 0
    assert status['locale'] == 'en-US'
    assert status['active_titles'] is not None
    assert len(status['active_titles']) == 1
    active_title = status['active_titles'][0]

    assert active_title['title_id'] == 714681658
    assert active_title['aum'] == 'AIVDE_s9eep9cpjhg6g!App'
    assert active_title['name'] == 'AIVDE_s9eep9cpjhg6g!App'
    assert active_title['image'] is None
    assert active_title['type'] is None
    assert active_title['has_focus'] is True
    assert active_title['title_location'] == 'StartView'
    assert active_title['product_id'] == '00000000-0000-0000-0000-000000000000'
    assert active_title['sandbox_id'] == '00000000-0000-0000-0000-000000000000'



from binascii import hexlify, unhexlify

from xbox.sg import packer, enum
from xbox.sg.constants import MessageTarget


def _unpack(data, crypto):
    return packer.unpack(data, crypto)


def _pack(msg, crypto):
    return packer.pack(msg, crypto)


def test_poweron_request(packets, crypto):
    unpacked = _unpack(packets['poweron_request'], crypto)

    assert unpacked.header.pkt_type == enum.PacketType.PowerOnRequest
    assert unpacked.header.unprotected_payload_length == 19
    assert unpacked.header.version == 0

    assert unpacked.unprotected_payload.liveid == 'FD00112233FFEE66'


def test_discovery_request(packets, crypto):
    unpacked = _unpack(packets['discovery_request'], crypto)

    assert unpacked.header.pkt_type == enum.PacketType.DiscoveryRequest
    assert unpacked.header.unprotected_payload_length == 10
    assert unpacked.header.version == 0

    assert unpacked.unprotected_payload.flags == 0
    assert unpacked.unprotected_payload.client_type == enum.ClientType.Android
    assert unpacked.unprotected_payload.minimum_version == 0
    assert unpacked.unprotected_payload.maximum_version == 2


def test_discovery_response(packets, crypto, public_key):
    unpacked = _unpack(packets['discovery_response'], crypto)

    assert unpacked.header.pkt_type == enum.PacketType.DiscoveryResponse
    assert unpacked.header.unprotected_payload_length == 580
    assert unpacked.header.version == 2
    assert unpacked.unprotected_payload.name == 'XboxOne'
    assert str(unpacked.unprotected_payload.uuid) == 'DE305D54-75B4-431B-ADB2-EB6B9E546014'.lower()
    assert unpacked.unprotected_payload.last_error == 0
    assert unpacked.unprotected_payload.cert.liveid == 'FFFFFFFFFFF'
    assert unpacked.unprotected_payload.cert.pubkey.public_numbers() == public_key.public_numbers()


def test_connect_request(packets, crypto):
    unpacked = _unpack(packets['connect_request'], crypto)
    assert unpacked.header.pkt_type == enum.PacketType.ConnectRequest
    assert unpacked.header.unprotected_payload_length == 98
    assert unpacked.header.protected_payload_length == 47
    assert unpacked.header.version == 2
    assert str(unpacked.unprotected_payload.sg_uuid) == 'de305d54-75b4-431b-adb2-eb6b9e546014'
    assert unpacked.unprotected_payload.public_key == b'\xff' * 64
    assert unpacked.unprotected_payload.iv == unhexlify('2979d25ea03d97f58f46930a288bf5d2')
    assert unpacked.protected_payload.userhash == 'deadbeefdeadbeefde'
    assert unpacked.protected_payload.jwt == 'dummy_token'
    assert unpacked.protected_payload.connect_request_num == 0
    assert unpacked.protected_payload.connect_request_group_start == 0
    assert unpacked.protected_payload.connect_request_group_end == 2


def test_connect_response(packets, crypto):
    unpacked = _unpack(packets['connect_response'], crypto)

    assert unpacked.header.pkt_type == enum.PacketType.ConnectResponse
    assert unpacked.header.unprotected_payload_length == 16
    assert unpacked.header.protected_payload_length == 8
    assert unpacked.header.version == 2
    assert unpacked.unprotected_payload.iv == unhexlify('c6373202bdfd1167cf9693491d22322a')
    assert unpacked.protected_payload.connect_result == enum.ConnectionResult.Success
    assert unpacked.protected_payload.pairing_state == enum.PairedIdentityState.NotPaired
    assert unpacked.protected_payload.participant_id == 31


def test_local_join(packets, crypto):
    unpacked = _unpack(packets['local_join'], crypto)
    payload = unpacked.protected_payload

    assert unpacked.header.flags.msg_type == enum.MessageType.LocalJoin
    assert payload.device_type == enum.ClientType.Android
    assert payload.native_width == 600
    assert payload.native_height == 1024
    assert payload.dpi_x == 160
    assert payload.dpi_y == 160
    assert payload.device_capabilities == enum.DeviceCapabilities.All
    assert payload.client_version == 133713371
    assert payload.os_major_version == 42
    assert payload.os_minor_version == 0
    assert payload.display_name == 'package.name.here'


def test_acknowledge(packets, crypto):
    unpacked = _unpack(packets['acknowledge'], crypto)
    payload = unpacked.protected_payload

    assert unpacked.header.flags.msg_type == enum.MessageType.Ack
    assert payload.low_watermark == 0
    assert payload.processed_list == [1]
    assert payload.rejected_list == []


def test_paired_identity_state_changed(packets, crypto):
    unpacked = _unpack(packets['paired_identity_state_changed'], crypto)
    payload = unpacked.protected_payload

    assert unpacked.header.flags.msg_type == enum.MessageType.PairedIdentityStateChanged
    assert payload.state == enum.PairedIdentityState.Paired


def test_fragmented_message(packets, crypto):
    data = unhexlify(
        b'6a5297cd00424d6963726f736f66742e426c75726179506c617965725f3877656b79623364386262'
        b'77652158626f782e426c75726179506c617965722e4170706c69636174696f6e0008883438303036'
        b'31303036433030364630303230303035343030363830303635303032303030343630303631303036'
        b'43303036433030323030303646303036363030323030303532303036353030363130303633303036'
        b'38303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'303030303030303030303030303030303030303030303030'
    )

    unpacked = _unpack(packets['fragment_media_state_0'], crypto)
    payload = unpacked.protected_payload

    assert unpacked.header.flags.msg_type == enum.MessageType.MediaState
    assert payload.sequence_begin == 22
    assert payload.sequence_end == 25
    assert payload.data == data


def test_gamedvr_record(packets, crypto):
    unpacked = _unpack(packets['gamedvr_record'], crypto)
    payload = unpacked.protected_payload

    assert unpacked.header.flags.msg_type == enum.MessageType.GameDvrRecord
    assert payload.start_time_delta == -60
    assert payload.end_time_delta == 0


def test_title_launch(packets, crypto):
    unpacked = _unpack(packets['title_launch'], crypto)
    payload = unpacked.protected_payload

    assert unpacked.header.flags.msg_type == enum.MessageType.TitleLaunch
    assert payload.location == enum.ActiveTitleLocation.Fill
    assert payload.uri == 'ms-xbl-0D174C79://default/'


def test_start_channel_request(packets, crypto):
    unpacked = _unpack(packets['start_channel_request'], crypto)
    payload = unpacked.protected_payload

    assert unpacked.header.flags.msg_type == enum.MessageType.StartChannelRequest
    assert payload.channel_request_id == 1
    assert payload.title_id == 0
    assert payload.service == MessageTarget.SystemInputUUID
    assert payload.activity_id == 0


def test_start_channel_request_title(packets, crypto):
    unpacked = _unpack(packets['start_channel_request_title_channel'], crypto)
    payload = unpacked.protected_payload

    assert unpacked.header.flags.msg_type == enum.MessageType.StartChannelRequest
    assert payload.channel_request_id == 16
    assert payload.title_id == 1256782258
    assert payload.service == MessageTarget.TitleUUID
    assert payload.activity_id == 0


def test_start_channel_response(packets, crypto):
    unpacked = _unpack(packets['start_channel_response'], crypto)
    payload = unpacked.protected_payload

    assert unpacked.header.flags.msg_type == enum.MessageType.StartChannelResponse
    assert payload.channel_request_id == 1
    assert payload.target_channel_id == 148
    assert payload.result == enum.SGResultCode.SG_E_SUCCESS


def test_console_status(packets, crypto):
    unpacked = _unpack(packets['console_status'], crypto)
    payload = unpacked.protected_payload

    assert unpacked.header.flags.msg_type == enum.MessageType.ConsoleStatus
    assert payload.live_tv_provider == 0
    assert payload.major_version == 10
    assert payload.minor_version == 0
    assert payload.build_number == 14393
    assert payload.locale == 'en-US'
    assert len(payload.active_titles) == 1
    assert payload.active_titles[0].title_id == 714681658
    assert payload.active_titles[0].disposition.has_focus is True
    assert payload.active_titles[0].disposition.title_location == enum.ActiveTitleLocation.StartView
    assert str(payload.active_titles[0].product_id) == '00000000-0000-0000-0000-000000000000'
    assert str(payload.active_titles[0].sandbox_id) == '00000000-0000-0000-0000-000000000000'
    assert payload.active_titles[0].aum == 'Xbox.Home_8wekyb3d8bbwe!Xbox.Home.Application'


def test_active_surface_change(packets, crypto):
    unpacked = _unpack(packets['active_surface_change'], crypto)
    payload = unpacked.protected_payload

    assert unpacked.header.flags.msg_type == enum.MessageType.ActiveSurfaceChange
    assert payload.surface_type == enum.ActiveSurfaceType.HTML
    assert payload.server_tcp_port == 0
    assert payload.server_udp_port == 0
    assert str(payload.session_id) == '00000000-0000-0000-0000-000000000000'
    assert payload.render_width == 0
    assert payload.render_height == 0
    assert hexlify(payload.master_session_key) == b'0000000000000000000000000000000000000000000000000000000000000000'


def test_auxiliary_stream_hello(packets, crypto):
    unpacked = _unpack(packets['auxiliary_stream_hello'], crypto)
    payload = unpacked.protected_payload

    assert unpacked.header.flags.msg_type == enum.MessageType.AuxilaryStream
    assert payload.connection_info_flag == 0


def test_auxiliary_stream_connection_info(packets, crypto):
    unpacked = _unpack(packets['auxiliary_stream_connection_info'], crypto)
    payload = unpacked.protected_payload

    assert unpacked.header.flags.msg_type == enum.MessageType.AuxilaryStream
    assert payload.connection_info_flag == 1
    assert hexlify(payload.connection_info.crypto_key) == b'14188d32cca3564a6d53f34ad8d21728'
    assert hexlify(payload.connection_info.server_iv) == b'09dcb570c9715cf01e0dfaf5ac718445'
    assert hexlify(payload.connection_info.client_iv) == b'9fa17a415b1bab5ae320cdceb5e37297'
    assert hexlify(payload.connection_info.sign_hash) == b'473f076d2fee90b27821fcad9d0ae7efdfd08f250823db95b90f90cac95784f9'
    assert len(payload.connection_info.endpoints) == 1
    assert payload.connection_info.endpoints[0].ip == '192.168.8.104'
    assert payload.connection_info.endpoints[0].port == '57344'


def test_disconnect(packets, crypto):
    unpacked = _unpack(packets['disconnect'], crypto)
    payload = unpacked.protected_payload

    assert unpacked.header.flags.msg_type == enum.MessageType.Disconnect
    assert payload.reason == enum.DisconnectReason.Unspecified
    assert payload.error_code == 0


def test_poweroff(packets, crypto):
    unpacked = _unpack(packets['power_off'], crypto)
    payload = unpacked.protected_payload

    assert unpacked.header.flags.msg_type == enum.MessageType.PowerOff
    assert payload.liveid == 'FD00112233FFEE66'


def test_json(packets, crypto):
    unpacked = _unpack(packets['json'], crypto)
    payload = unpacked.protected_payload

    assert unpacked.header.flags.msg_type == enum.MessageType.Json
    assert payload.text == {'request': 'GetConfiguration', 'msgid': '2ed6c0fd.2'}


def test_gamepad(packets, crypto):
    unpacked = _unpack(packets['gamepad'], crypto)
    payload = unpacked.protected_payload

    assert unpacked.header.flags.msg_type == enum.MessageType.Gamepad
    assert payload.timestamp == 0
    assert payload.buttons == enum.GamePadButton.PadB
    assert payload.left_trigger == 0.0
    assert payload.right_trigger == 0.0
    assert payload.left_thumbstick_x == 0.0
    assert payload.left_thumbstick_y == 0.0
    assert payload.right_thumbstick_x == 0.0
    assert payload.right_thumbstick_y == 0.0


def test_media_command(packets, crypto):
    unpacked = _unpack(packets['media_command'], crypto)
    payload = unpacked.protected_payload

    assert unpacked.header.flags.msg_type == enum.MessageType.MediaCommand
    assert payload.request_id == 0
    assert payload.title_id == 274278798
    assert payload.command == enum.MediaControlCommand.FastForward
    assert payload.seek_position is None


def test_media_state(packets, crypto):
    unpacked = _unpack(packets['media_state'], crypto)
    payload = unpacked.protected_payload

    assert unpacked.header.flags.msg_type == enum.MessageType.MediaState
    assert payload.title_id == 274278798
    assert payload.aum_id == 'AIVDE_s9eep9cpjhg6g!App'
    assert payload.asset_id == ''
    assert payload.media_type == enum.MediaType.NoMedia
    assert payload.sound_level == enum.SoundLevel.Full
    assert payload.enabled_commands == enum.MediaControlCommand(33758)
    assert payload.playback_status == enum.MediaPlaybackStatus.Stopped
    assert payload.rate == 0.0
    assert payload.position == 0
    assert payload.media_start == 0
    assert payload.media_end == 0
    assert payload.min_seek == 0
    assert payload.max_seek == 0
    assert len(payload.metadata) == 1
    assert payload.metadata[0].name == 'title'
    assert payload.metadata[0].value == ''


def test_text_acknowledge(packets, crypto):
    unpacked = _unpack(packets['system_text_acknowledge'], crypto)
    payload = unpacked.protected_payload

    assert unpacked.header.flags.msg_type == enum.MessageType.SystemTextAck
    assert payload.text_session_id == 8
    assert payload.text_version_ack == 2


def test_text_configuration(packets, crypto):
    unpacked = _unpack(packets['system_text_configuration'], crypto)
    payload = unpacked.protected_payload

    assert unpacked.header.flags.msg_type == enum.MessageType.SystemTextConfiguration
    assert payload.text_session_id == 9
    assert payload.text_buffer_version == 0
    assert payload.text_options == enum.TextOption.AcceptsReturn | enum.TextOption.MultiLine
    assert payload.input_scope == enum.TextInputScope.Unknown
    assert payload.max_text_length == 0
    assert payload.locale == 'de-DE'
    assert payload.prompt == ''


def test_system_text_done(packets, crypto):
    unpacked = _unpack(packets['system_text_done'], crypto)
    payload = unpacked.protected_payload

    assert unpacked.header.flags.msg_type == enum.MessageType.SystemTextDone
    assert payload.text_session_id == 0
    assert payload.text_version == 0
    assert payload.flags == 0
    assert payload.result == enum.TextResult.Cancel


def test_system_text_input(packets, crypto):
    unpacked = _unpack(packets['system_text_input'], crypto)
    payload = unpacked.protected_payload

    assert unpacked.header.flags.msg_type == enum.MessageType.SystemTextInput
    assert payload.text_session_id == 8
    assert payload.base_version == 1
    assert payload.submitted_version == 2
    assert payload.total_text_byte_len == 1
    assert payload.selection_start == -1
    assert payload.selection_length == -1
    assert payload.flags == 0
    assert payload.text_chunk_byte_start == 0
    assert payload.delta is None
    assert payload.text_chunk == 'h'


def test_system_touch(packets, crypto):
    unpacked = _unpack(packets['system_touch'], crypto)
    payload = unpacked.protected_payload

    assert unpacked.header.flags.msg_type == enum.MessageType.SystemTouch
    assert payload.touch_msg_timestamp == 182459592
    assert len(payload.touchpoints) == 1
    assert payload.touchpoints[0].touchpoint_id == 1
    assert payload.touchpoints[0].touchpoint_action == enum.TouchAction.Down
    assert payload.touchpoints[0].touchpoint_x == 244
    assert payload.touchpoints[0].touchpoint_y == 255


def test_repack_all(packets, crypto):
    for f in packets:
        unpacked = _unpack(packets[f], crypto)

        # if unpacked.header.pkt_type in simple.pkt_types:
        #     msg = simple.struct(**unpacked)
        # elif unpacked.header.pkt_type == enum.PacketType.Message and unpacked.header.flags.msg_type == enum.MessageType.Json:
        #     continue  # The key order is not preserved when parsing JSON so the parsed message will always differ
        # else:
        #     msg = message.struct(**unpacked)

        repacked = _pack(unpacked, crypto)

        assert repacked == packets[f], \
            '%s was not repacked correctly:\n(repacked)%s\n!=\n(original)%s'\
            % (f, hexlify(repacked), hexlify(packets[f]))



"""
Constant values used by SmartGlass
"""
from uuid import UUID

from xbox.sg.enum import ClientType, DeviceCapabilities


class AndroidClientInfo(object):
    """
    Client Info for Android device (tablet). Used for LocalJoin messages
    """
    DeviceType = ClientType.Android
    # Resolution is portrait mode
    NativeWidth = 720
    NativeHeight = 1280
    DpiX = 160
    DpiY = 160
    DeviceCapabilities = DeviceCapabilities.All
    ClientVersion = 151117100  # v2.4.1511.17100-Beta
    OSMajor = 22  # Android 5.1.1 - API Version 22
    OSMinor = 0
    DisplayName = "com.microsoft.xboxone.smartglass.beta"


class WindowsClientInfo(object):
    """
    Client Info for Windows device, used for LocalJoin messages
    """
    DeviceType = ClientType.WindowsStore
    NativeWidth = 1080
    NativeHeight = 1920
    DpiX = 96
    DpiY = 96
    DeviceCapabilities = DeviceCapabilities.All
    ClientVersion = 39
    OSMajor = 6
    OSMinor = 2
    DisplayName = "SmartGlass-PC"


class MessageTarget(object):
    """
    UUIDs for all ServiceChannels
    """
    SystemInputUUID = UUID("fa20b8ca-66fb-46e0-adb60b978a59d35f")
    SystemInputTVRemoteUUID = UUID("d451e3b3-60bb-4c71-b3dbf994b1aca3a7")
    SystemMediaUUID = UUID("48a9ca24-eb6d-4e12-8c43d57469edd3cd")
    SystemTextUUID = UUID("7af3e6a2-488b-40cb-a93179c04b7da3a0")
    SystemBroadcastUUID = UUID("b6a117d8-f5e2-45d7-862e8fd8e3156476")
    TitleUUID = UUID('00000000-0000-0000-0000-000000000000')


class XboxOneGuid(object):
    """
    System level GUIDs
    """
    BROWSER = "9c7e0f20-78fb-4ea7-a8bd-cf9d78059a08"
    MUSIC = "6D96DEDC-F3C9-43F8-89E3-0C95BF76AD2A"
    VIDEO = "a489d977-8a87-4983-8df6-facea1ad6d93"


class TitleId(object):
    """
    System level Title Ids
    """
    AVATAR_EDITOR_TITLE_ID = 1481443281
    BROWSER_TITLE_ID = 1481115776
    DASH_TITLE_ID = 4294838225
    BLUERAY_TITLE_ID = 1783797709
    CONSOLE_UPDATE_APP_TITLE_ID = -1
    HOME_TITLE_ID = 714681658
    IE_TITLE_ID = 1032557327
    LIVETV_TITLE_ID = 371594669
    MUSIC_TITLE_ID = 419416564
    ONEGUIDE_TITLE_ID = 2055214557
    OOBE_APP_TITLE_ID = 951105730
    SNAP_AN_APP_TITLE_ID = 1783889234
    STORE_APP_TITLE_ID = 1407783715
    VIDEO_TITLE_ID = 1030770725
    ZUNE_TITLE_ID = 1481115739



"""
Smartglass packet factory
"""
from construct import Container
from xbox.sg.enum import PacketType, MessageType, ClientType
from xbox.sg.packet import simple, message

CHANNEL_CORE = 0


def _message_header(msg_type, channel_id=0, target_participant_id=0,
                    source_participant_id=0, is_fragment=False,
                    need_ack=False):
    """
    Helper method for creating a message header.

    Args:
        msg_type (int): The message type.
        channel_id (int): The target channel of the message.
        target_participant_id (int): The target participant Id.
        source_participant_id (int): The source participant Id.
        is_fragment (bool): Whether the message is a fragment.
        need_ack (bool): Whether the message needs an acknowledge.

    Returns:
        :class:`XStructObj`: Instance of :class:`:class:`XStructObj``.
    """
    return message.header(
        sequence_number=0,
        target_participant_id=target_participant_id,
        source_participant_id=source_participant_id,
        flags=Container(
            need_ack=need_ack,
            is_fragment=is_fragment,
            msg_type=msg_type
        ),
        channel_id=channel_id,
        pkt_type=PacketType.Message
    )


def power_on(liveid):
    """
    Assemble PowerOn Request message.

    Args:
        liveid (str): The console LiveId (extracted from the
                      :class:`DiscoveryResponse` Certificate).

    Returns:
        :class:`XStructObj`: Instance of :class:`:class:`XStructObj``.
    """
    return simple.struct(
        header=simple.header(pkt_type=PacketType.PowerOnRequest, version=0),
        unprotected_payload=simple.power_on_request(liveid=liveid)
    )


def discovery(client_type=ClientType.Android):
    """
    Assemble DiscoveryRequest SimpleMessage.

    Args:
        client_type (:class:`ClientType`): Member of :class:`ClientType`, defaults to
                           `ClientType.Android`.

    Returns:
        :class:`XStructObj`: Instance of :class:`:class:`XStructObj``.
    """
    return simple.struct(
        header=simple.header(pkt_type=PacketType.DiscoveryRequest, version=0),
        unprotected_payload=simple.discovery_request(
            flags=0, client_type=client_type,
            minimum_version=0, maximum_version=2
        )
    )


def connect(sg_uuid, public_key_type, public_key, iv, userhash, jwt,
            msg_num, num_start, num_end):
    """
    Assemble ConnectRequest SimpleMessage.

    Args:
        sg_uuid (UUID): Client Uuid, randomly generated.
        public_key_type (:class:`PublicKeyType`): Public Key Type.
        public_key (bytes): Calculated Public Key, from :class:`Crypto`.
        iv (bytes): Initialization Vector for this encrypted message.
        userhash (str): Xbox Live Account userhash.
        jwt (str): Xbox Live Account JWT / Auth-token.
        msg_num (int): Current message number (important for fragmentation).
        num_start (int): Base number start of ConnectRequest fragments.
        num_end (int): Base number end of ConnectRequest fragments
                       (base number start + total fragments + 1).

    Returns:
        :class:`XStructObj`: Instance of :class:`:class:`XStructObj``.
    """
    return simple.struct(
        header=simple.header(pkt_type=PacketType.ConnectRequest),
        unprotected_payload=simple.connect_request_unprotected(
            sg_uuid=sg_uuid, public_key_type=public_key_type,
            public_key=public_key, iv=iv
        ),
        protected_payload=simple.connect_request_protected(
            userhash=userhash, jwt=jwt, connect_request_num=msg_num,
            connect_request_group_start=num_start,
            connect_request_group_end=num_end
        )
    )


def message_fragment(msg_type, sequence_begin, sequence_end, data, **kwargs):
    """
    Assemble fragmented message.

    Args:
        msg_type (int): Base Message Type.
        sequence_begin (int): Sequence number with first fragment.
        sequence_end (int): Last sequence number (+1) containing fragment.
        data (bytes): Plaintext MessagePacket payload fragment.

    Returns:
        :class:`XStructObj`: Instance of :class:`:class:`XStructObj``.
    """
    return message.struct(
        header=_message_header(
            msg_type, is_fragment=True, **kwargs
        ),
        protected_payload=message.fragment(
            sequence_begin=sequence_begin,
            sequence_end=sequence_end,
            data=data
        )
    )


def acknowledge(low_watermark, processed_list, rejected_list, **kwargs):
    """
    Assemble acknowledgement message.

    Args:
        low_watermark (int): Low Watermark.
        processed_list (list): List of processed message sequence numbers.
        rejected_list (list): List of rejected message sequence numbers.

    Returns:
        :class:`XStructObj`: Instance of :class:`:class:`XStructObj``.
    """
    return message.struct(
        header=_message_header(
            MessageType.Ack, **kwargs
        ),
        protected_payload=message.acknowledge(
            low_watermark=low_watermark,
            processed_list=processed_list,
            rejected_list=rejected_list
        )
    )


def json(text, **kwargs):
    """
    Assemble JSON message.

    Args:
        text (str): Text string.

    Returns:
        :class:`XStructObj`: Instance of :class:`:class:`XStructObj``.
    """
    return message.struct(
        header=_message_header(
            MessageType.Json,
            need_ack=True, **kwargs
        ),
        protected_payload=message.json(
            text=text
        )
    )


def disconnect(reason, error_code, **kwargs):
    """
    Assemble Disconnect message.

    Args:
        reason (:class:`xbox.sg.enum.DisconnectReason`): Disconnect reason.
        error_code (int): Error code.

    Returns:
        :class:`XStructObj`: Instance of :class:`:class:`XStructObj``.
    """
    return message.struct(
        header=_message_header(
            MessageType.Disconnect, CHANNEL_CORE, **kwargs
        ),
        protected_payload=message.disconnect(
            reason=reason,
            error_code=error_code
        )
    )


def power_off(liveid, **kwargs):
    """
    Assemble PowerOff message.

    Args:
        liveid (str): Live ID of console.

    Returns:
        :class:`XStructObj`: Instance of :class:`:class:`XStructObj``.
    """
    return message.struct(
        header=_message_header(
            MessageType.PowerOff, CHANNEL_CORE, **kwargs
        ),
        protected_payload=message.power_off(
            liveid=liveid
        )
    )


def local_join(client_info, **kwargs):
    """
    Assemble LocalJoin message.

    Args:
        client_info (object): Instance of :class:`WindowsClientInfo`
                              or :class:`AndroidClientInfo`.

    Returns:
        :class:`XStructObj`: Instance of :class:`:class:`XStructObj``.
    """
    return message.struct(
        header=_message_header(
            MessageType.LocalJoin, CHANNEL_CORE, need_ack=True, **kwargs
        ),
        protected_payload=message.local_join(
            device_type=client_info.DeviceType,
            native_width=client_info.NativeWidth,
            native_height=client_info.NativeHeight,
            dpi_x=client_info.DpiX, dpi_y=client_info.DpiY,
            device_capabilities=client_info.DeviceCapabilities,
            client_version=client_info.ClientVersion,
            os_major_version=client_info.OSMajor,
            os_minor_version=client_info.OSMinor,
            display_name=client_info.DisplayName
        )
    )


def title_launch(location, uri, **kwargs):
    """
    Assemble TitleLaunch message.

    Args:
        location (:class:`ActiveTitleLocation`): Location.
        uri (str): Uri string for title to launch.

    Returns:
        :class:`XStructObj`: Instance of :class:`:class:`XStructObj``.
    """
    return message.struct(
        header=_message_header(
            MessageType.TitleLaunch, CHANNEL_CORE, **kwargs
        ),
        protected_payload=message.title_launch(
            location=location,
            uri=uri
        )
    )


def start_channel(channel_request_id, title_id, service, activity_id,
                  **kwargs):
    """
    Assemble StartChannelRequest message.

    Args:
        channel_request_id (int): Incrementing Channel Request Id.
        title_id (int): Title Id, usually 0.
        service (:class:`MessageTarget`): Member of :class:`MessageTarget`.
        activity_id (int): Activity Id, usually 0.

    Returns:
        :class:`XStructObj`: Instance of :class:`:class:`XStructObj``.
    """
    return message.struct(
        header=_message_header(
            MessageType.StartChannelRequest, CHANNEL_CORE,
            need_ack=True, **kwargs
        ),
        protected_payload=message.start_channel_request(
            channel_request_id=channel_request_id,
            title_id=title_id,
            service=service,
            activity_id=activity_id
        )
    )


def stop_channel(channel_id, **kwargs):
    """
    Assemble StopChannel message.

    Args:
        channel_id (int): Channel Id to stop.

    Returns:
        :class:`XStructObj`: Instance of :class:`:class:`XStructObj``.
    """
    return message.struct(
        header=_message_header(
            MessageType.StopChannel, CHANNEL_CORE, **kwargs
        ),
        protected_payload=message.stop_channel(
            target_channel_id=channel_id
        )
    )


def gamepad(timestamp, buttons, l_trigger, r_trigger, l_thumb_x, l_thumb_y,
            r_thumb_x, r_thumb_y, **kwargs):
    """
    Assemble gamepad input message.

    Args:
        timestamp (longlong): Timestamp.
        buttons (:class:`GamePadButton`): Bitmask of pressed gamepad buttons.
        l_trigger (float): LT.
        r_trigger (float): RT.
        l_thumb_x (float): Position of left thumbstick, X-Axis.
        l_thumb_y (float): Position of left thumbstick, Y-Axis.
        r_thumb_x (float): Position of right thumbstick, X-Axis.
        r_thumb_y (float): Position of right thumbstick, Y-Axis.

    Returns:
        :class:`XStructObj`: Instance of :class:`:class:`XStructObj``.
    """
    return message.struct(
        header=_message_header(
            MessageType.Gamepad, **kwargs
        ),
        protected_payload=message.gamepad(
            timestamp=timestamp, buttons=buttons,
            left_trigger=l_trigger, right_trigger=r_trigger,
            left_thumbstick_x=l_thumb_x, left_thumbstick_y=l_thumb_y,
            right_thumbstick_x=r_thumb_x, right_thumbstick_y=r_thumb_y
        )
    )


def unsnap(unknown, **kwargs):
    """
    Assemble unsnap message.

    Args:
        unknown (int): Unknown value.

    Returns:
        :class:`XStructObj`: Instance of :class:`:class:`XStructObj``.
    """
    return message.struct(
        header=_message_header(
            MessageType.Unsnap, CHANNEL_CORE, **kwargs
        ),
        protected_payload=message.unsnap(
            unk=unknown
        )
    )


def game_dvr_record(start_time_delta, end_time_delta, **kwargs):
    """
    Assemble Game DVR record message.

    Args:
        start_time_delta (int): Start Time delta.
        end_time_delta (int): End Time delta.

    Returns:
        :class:`XStructObj`: Instance of :class:`:class:`XStructObj``.
    """
    return message.struct(
        header=_message_header(
            MessageType.GameDvrRecord, CHANNEL_CORE,
            need_ack=True, **kwargs
        ),
        protected_payload=message.game_dvr_record(
            start_time_delta=start_time_delta,
            end_time_delta=end_time_delta
        )
    )


def media_command(request_id, title_id, command, seek_position, **kwargs):
    """
    Assemble Media Command message.

    Args:
        request_id (int): Request Id of MediaCommand.
        title_id (int): Title Id of Application to control.
        command (:class:`MediaControlCommand`): Media Command.
        seek_position (ulonglong): Seek position.

     Returns:
        :class:`XStructObj`: Instance of :class:`:class:`XStructObj``.
    """
    return message.struct(
        header=_message_header(
            MessageType.MediaCommand,
            need_ack=True, **kwargs
        ),
        protected_payload=message.media_command(
            request_id=request_id,
            title_id=title_id,
            command=command,
            seek_position=seek_position
        )
    )


def systemtext_input(session_id, base_version, submitted_version, total_text_len,
                     selection_start, selection_length, flags, text_chunk_byte_start,
                     text_chunk, delta=None, **kwargs):
    """
    Assemble SystemText Input message

    Args:
        session_id (int): Textt session Id
        base_version (int): Base version
        submitted_version (int): Submitted Version
        total_text_len (int): Total text length
        selection_start (int): Selection start
        selection_length (int): Selection length
        flags (int): Input flags
        text_chunk_byte_start (int): Start byte of text chunk
        text_chunk (str): Actual text to send
        delta (NoneType): Unknown

    Returns:
        :class:`XStructObj`: Instance of :class:`:class:`XStructObj``.
    """
    return message.struct(
        header=_message_header(
            MessageType.SystemTextInput,
            need_ack=True, **kwargs
        ),
        protected_payload=message.system_text_input(
            text_session_id=session_id,
            base_version=base_version,
            submitted_version=submitted_version,
            total_text_byte_len=total_text_len,
            selection_start=selection_start,
            selection_length=selection_length,
            flags=flags,
            text_chunk_byte_start=text_chunk_byte_start,
            text_chunk=text_chunk
            # delta=delta
        )
    )


def systemtext_ack(session_id, text_version, **kwargs):
    """
    Assemble SystemText Acknowledge message

    Args:
        session_id (int): Text session Id
        text_version (int): Text version to acknowledge

    Returns:
        :class:`XStructObj`: Instance of :class:`:class:`XStructObj``.
    """
    return message.struct(
        header=_message_header(
            MessageType.SystemTextAck,
            need_ack=True, **kwargs
        ),
        protected_payload=message.system_text_acknowledge(
            text_session_id=session_id,
            text_version_ack=text_version
        )
    )


def systemtext_done(session_id, text_version, flags, result, **kwargs):
    """
    Assemble SystemText Done message

    Args:
        session_id (int): Text session Id
        text_version (int): Text version
        flags (int): Flags
        result (:class:`TextResult`): Text result

    Returns:
        :class:`XStructObj`: Instance of :class:`:class:`XStructObj``.
    """
    return message.struct(
        header=_message_header(
            MessageType.SystemTextDone,
            need_ack=True, **kwargs
        ),
        protected_payload=message.system_text_done(
            text_session_id=session_id,
            text_version=text_version,
            flags=flags,
            result=result
        )
    )


def title_auxiliary_stream(**kwargs):
    """
    Assemble Auxiliary Stream message

    Returns:
         :class:`XStructObj`: Instance of :class:`XStructObj`
    """
    return message.struct(
        header=_message_header(
            MessageType.AuxilaryStream,
            need_ack=True, **kwargs
        ),
        protected_payload=message.auxiliary_stream(
            connection_info_flag=0
        )
    )



"""
Input smartglass client

Send controller input via stdin (terminal) to the console
"""
import sys
import logging

import gevent.socket

from xbox.sg.enum import GamePadButton

LOGGER = logging.getLogger(__name__)

input_map = {
    "i": GamePadButton.DPadUp,
    "k": GamePadButton.DPadDown,
    "j": GamePadButton.DPadLeft,
    "l": GamePadButton.DPadRight,

    "a": GamePadButton.PadA,
    "b": GamePadButton.PadB,
    "x": GamePadButton.PadX,
    "y": GamePadButton.PadY,

    "t": GamePadButton.View,
    "z": GamePadButton.Nexu,
    "u": GamePadButton.Menu
}


def get_getch_func():
    """
    Source: https://code.activestate.com/recipes/577977-get-single-keypress/
    """
    try:
        import tty
        import termios
    except ImportError:
        # Probably Windows.
        try:
            import msvcrt
        except ImportError:
            # Just give up here.
            raise ImportError('getch not available')
        else:
            return msvcrt.getch
    else:
        def getch():
            """
            getch() -> key character

            Read a single keypress from stdin and return the resulting character.
            Nothing is echoed to the console. This call will block if a keypress
            is not already available, but will not wait for Enter to be pressed.

            If the pressed key was a modifier key, nothing will be detected; if
            it were a special function key, it may return the first character of
            of an escape sequence, leaving additional characters in the buffer.
            """
            fd = sys.stdin.fileno()
            old_settings = termios.tcgetattr(fd)
            try:
                tty.setraw(fd)
                gevent.socket.wait_read(fd)
                ch = sys.stdin.read(1)
            finally:
                termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
            return ch
        return getch


def input_loop(console):
    getch = get_getch_func()
    while True:
        ch = getch()
        print(ch)
        if ord(ch) == 3:  # CTRL-C
            sys.exit(1)

        elif ch not in input_map:
            continue

        button = input_map[ch]
        console.gamepad_input(button)
        console.wait(0.1)
        console.gamepad_input(GamePadButton.Clear)



import pytest

from xbox.rest.consolewrap import ConsoleWrap
from xbox.sg import enum


def test_consolewrap_init(console):
    wrap = ConsoleWrap(console)

    assert wrap.console == console
    assert 'text' in wrap.console.managers
    assert 'input' in wrap.console.managers
    assert 'stump' in wrap.console.managers
    assert 'media' in wrap.console.managers


def test_discover():
    discovered = ConsoleWrap.discover(tries=1, blocking=False, timeout=1)

    assert isinstance(discovered, list)


def test_poweron():
    ConsoleWrap.power_on('FD0123456789', tries=1, iterations=1)


def test_media_commands(console):
    commands = ConsoleWrap(console).media_commands

    assert isinstance(commands, dict)
    assert 'play' in commands
    for k, v in commands.items():
        assert isinstance(k, str)
        assert isinstance(v, enum.MediaControlCommand)


def test_input_keys(console):
    keys = ConsoleWrap(console).input_keys

    assert isinstance(keys, dict)
    assert 'nexus' in keys
    for k, v in keys.items():
        assert isinstance(k, str)
        assert isinstance(v, enum.GamePadButton)


def test_liveid(console):
    assert ConsoleWrap(console).liveid == console.liveid


def test_available(console):
    console._device_status = enum.DeviceStatus.Unavailable
    assert ConsoleWrap(console).available is False

    console._device_status = enum.DeviceStatus.Available
    assert ConsoleWrap(console).available is True


def test_connected(console):
    console._connection_state = enum.ConnectionState.Disconnected
    assert ConsoleWrap(console).connected is False

    console._connection_state = enum.ConnectionState.Connected
    assert ConsoleWrap(console).connected is True

    console._connection_state = enum.ConnectionState.Connecting
    assert ConsoleWrap(console).connected is False


def test_usable(console):
    console._connection_state = enum.ConnectionState.Reconnecting
    assert ConsoleWrap(console).usable is False

    console._connection_state = enum.ConnectionState.Disconnecting
    assert ConsoleWrap(console).usable is False

    console._connection_state = enum.ConnectionState.Disconnected
    assert ConsoleWrap(console).usable is False

    console._connection_state = enum.ConnectionState.Error
    assert ConsoleWrap(console).usable is False

    console._connection_state = enum.ConnectionState.Connected
    assert ConsoleWrap(console).usable is True


def test_connection_state(console):
    console._connection_state = enum.ConnectionState.Reconnecting

    assert ConsoleWrap(console).connection_state == enum.ConnectionState.Reconnecting


def test_pairing_state(console):
    console._pairing_state = enum.PairedIdentityState.Paired

    assert ConsoleWrap(console).pairing_state == enum.PairedIdentityState.Paired


def test_device_status(console):
    console._device_status = enum.DeviceStatus.Available

    assert ConsoleWrap(console).device_status == enum.DeviceStatus.Available


def test_authenticated_users_allowed(console):
    console.flags = enum.PrimaryDeviceFlag.AllowAuthenticatedUsers
    assert ConsoleWrap(console).authenticated_users_allowed is True

    console.flags = enum.PrimaryDeviceFlag.AllowConsoleUsers
    assert ConsoleWrap(console).authenticated_users_allowed is False

    console.flags = enum.PrimaryDeviceFlag.CertificatePending | enum.PrimaryDeviceFlag.AllowConsoleUsers
    assert ConsoleWrap(console).authenticated_users_allowed is False


def test_console_users_allowed(console):
    console.flags = enum.PrimaryDeviceFlag.AllowAuthenticatedUsers
    assert ConsoleWrap(console).console_users_allowed is False

    console.flags = enum.PrimaryDeviceFlag.AllowConsoleUsers
    assert ConsoleWrap(console).console_users_allowed is True

    console.flags = enum.PrimaryDeviceFlag.CertificatePending | enum.PrimaryDeviceFlag.AllowConsoleUsers
    assert ConsoleWrap(console).console_users_allowed is True


def test_anonymous_connection_allowed(console):
    console.flags = enum.PrimaryDeviceFlag.AllowAuthenticatedUsers
    assert ConsoleWrap(console).anonymous_connection_allowed is False

    console.flags = enum.PrimaryDeviceFlag.AllowConsoleUsers
    assert ConsoleWrap(console).anonymous_connection_allowed is False

    console.flags = enum.PrimaryDeviceFlag.CertificatePending | enum.PrimaryDeviceFlag.AllowAnonymousUsers
    assert ConsoleWrap(console).anonymous_connection_allowed is True


def test_is_certificate_pending(console):
    console.flags = enum.PrimaryDeviceFlag.AllowAuthenticatedUsers
    assert ConsoleWrap(console).is_certificate_pending is False

    console.flags = enum.PrimaryDeviceFlag.AllowConsoleUsers
    assert ConsoleWrap(console).is_certificate_pending is False

    console.flags = enum.PrimaryDeviceFlag.CertificatePending | enum.PrimaryDeviceFlag.AllowConsoleUsers
    assert ConsoleWrap(console).is_certificate_pending is True


def test_console_status(console, console_status):
    console._console_status = None
    assert ConsoleWrap(console).console_status is None

    console._console_status = console_status
    status = ConsoleWrap(console).console_status
    assert status is not None
    assert 'live_tv_provider' in status
    assert 'kernel_version' in status
    assert 'locale' in status
    assert 'active_titles' in status


def test_media_status(console, media_state, console_status, console_status_with_media):
    console.media._media_state = None
    assert ConsoleWrap(console).media_status is None

    console.media._media_state = media_state
    console._console_status = console_status_with_media
    console._connection_state = enum.ConnectionState.Disconnecting
    assert ConsoleWrap(console).media_status is None

    console._console_status = console_status  # miss-matched apps
    console._connection_state = enum.ConnectionState.Connected
    state = ConsoleWrap(console).media_status
    assert ConsoleWrap(console).media_status is None

    console._console_status = console_status_with_media
    state = ConsoleWrap(console).media_status
    assert 'title_id' in state
    assert 'aum_id' in state
    assert 'asset_id' in state
    assert 'media_type' in state
    assert 'sound_level' in state
    assert 'enabled_commands' in state
    assert 'playback_status' in state
    assert 'rate' in state
    assert 'position' in state
    assert 'media_start' in state
    assert 'media_end' in state
    assert 'min_seek' in state
    assert 'max_seek' in state
    assert 'metadata' in state


def test_status(console):
    status = ConsoleWrap(console).status
    assert status is not None

    assert isinstance(status['address'], str)
    assert isinstance(status['name'], str)
    assert isinstance(status['uuid'], str)
    assert isinstance(status['liveid'], str)

    assert isinstance(status['connection_state'], str)
    assert isinstance(status['pairing_state'], str)
    assert isinstance(status['device_status'], str)
    assert isinstance(status['authenticated_users_allowed'], bool)
    assert isinstance(status['console_users_allowed'], bool)
    assert isinstance(status['anonymous_connection_allowed'], bool)
    assert isinstance(status['is_certificate_pending'], bool)


def test_connect(console):
    console.flags = enum.PrimaryDeviceFlag.AllowAuthenticatedUsers
    console._device_status = enum.DeviceStatus.Available
    console._connection_state = enum.ConnectionState.Disconnected
    console._pairing_state = enum.PairedIdentityState.NotPaired

    with pytest.raises(Exception):
        ConsoleWrap(console).connect()

    console._connection_state = enum.ConnectionState.Connected
    state = ConsoleWrap(console).connect()
    assert state == enum.ConnectionState.Connected

    console._connection_state = enum.ConnectionState.Disconnected
    console.flags = enum.PrimaryDeviceFlag.AllowAnonymousUsers
    # blocks forever
    # state = ConsoleWrap(console).connect()
    # assert state == enum.ConnectionState.Disconnected



import logging

from xbox.sg import enum
from xbox.sg.console import Console
from xbox.sg.manager import InputManager, TextManager, MediaManager
from xbox.stump.manager import StumpManager

log = logging.getLogger()


class ConsoleWrap(object):
    def __init__(self, console):
        self.console = console

        if 'input' not in self.console.managers:
            self.console.add_manager(InputManager)
        if 'text' not in self.console.managers:
            self.console.add_manager(TextManager)
        if 'media' not in self.console.managers:
            self.console.add_manager(MediaManager)
        if 'stump' not in self.console.managers:
            self.console.add_manager(StumpManager)

        try:
            from xbox.nano.manager import NanoManager
            if 'nano' not in self.console.managers:
                self.console.add_manager(NanoManager)
        except ImportError:
            log.warning(
                'Failed to import NanoManager (depends on xbox-smartglass-nano).'
                ' /nano endpoint will not work!'
            )

    @staticmethod
    def discover(*args, **kwargs):
        return Console.discover(*args, **kwargs)

    @staticmethod
    def power_on(liveid, addr=None, iterations=3, tries=10):
        for i in range(iterations):
            Console.power_on(liveid, addr=addr, tries=tries)
            Console.wait(1)

    @property
    def media_commands(self):
        return {
            'play': enum.MediaControlCommand.Play,
            'pause': enum.MediaControlCommand.Pause,
            'play_pause': enum.MediaControlCommand.PlayPauseToggle,
            'stop': enum.MediaControlCommand.Stop,
            'record': enum.MediaControlCommand.Record,
            'next_track': enum.MediaControlCommand.NextTrack,
            'prev_track': enum.MediaControlCommand.PreviousTrack,
            'fast_forward': enum.MediaControlCommand.FastForward,
            'rewind': enum.MediaControlCommand.Rewind,
            'channel_up': enum.MediaControlCommand.ChannelUp,
            'channel_down': enum.MediaControlCommand.ChannelDown,
            'back': enum.MediaControlCommand.Back,
            'view': enum.MediaControlCommand.View,
            'menu': enum.MediaControlCommand.Menu,
            'seek': enum.MediaControlCommand.Seek
        }

    @property
    def input_keys(self):
        return {
            'clear': enum.GamePadButton.Clear,
            'enroll': enum.GamePadButton.Enroll,
            'nexus': enum.GamePadButton.Nexu,
            'menu': enum.GamePadButton.Menu,
            'view': enum.GamePadButton.View,
            'a': enum.GamePadButton.PadA,
            'b': enum.GamePadButton.PadB,
            'x': enum.GamePadButton.PadX,
            'y': enum.GamePadButton.PadY,
            'dpad_up': enum.GamePadButton.DPadUp,
            'dpad_down': enum.GamePadButton.DPadDown,
            'dpad_left': enum.GamePadButton.DPadLeft,
            'dpad_right': enum.GamePadButton.DPadRight,
            'left_shoulder': enum.GamePadButton.LeftShoulder,
            'right_shoulder': enum.GamePadButton.RightShoulder,
            'left_thumbstick': enum.GamePadButton.LeftThumbStick,
            'right_thumbstick': enum.GamePadButton.RightThumbStick
        }

    @property
    def liveid(self):
        return self.console.liveid

    @property
    def last_error(self):
        return self.console.last_error

    @property
    def available(self):
        return bool(self.console and self.console.available)

    @property
    def connected(self):
        return bool(self.console and self.console.connected)

    @property
    def usable(self):
        return bool(self.console and self.connected)

    @property
    def connection_state(self):
        if not self.console:
            return enum.ConnectionState.Disconnected

        return self.console.connection_state

    @property
    def pairing_state(self):
        if not self.console:
            return enum.PairedIdentityState.NotPaired

        return self.console.pairing_state

    @property
    def device_status(self):
        if not self.console:
            return enum.DeviceStatus.Unavailable

        return self.console.device_status

    @property
    def authenticated_users_allowed(self):
        return bool(self.console and self.console.authenticated_users_allowed)

    @property
    def console_users_allowed(self):
        return bool(self.console and self.console.console_users_allowed)

    @property
    def anonymous_connection_allowed(self):
        return bool(self.console and self.console.anonymous_connection_allowed)

    @property
    def is_certificate_pending(self):
        return bool(self.console and self.console.is_certificate_pending)

    @property
    def console_status(self):
        status_json = {}

        if not self.console or not self.console.console_status:
            return None

        status = self.console.console_status
        kernel_version = '{0}.{1}.{2}'.format(status.major_version, status.minor_version, status.build_number)

        status_json.update({
            'live_tv_provider': status.live_tv_provider,
            'kernel_version': kernel_version,
            'locale': status.locale
        })

        active_titles = []
        for at in status.active_titles:
            title = {
                'title_id': at.title_id,
                'aum': at.aum,
                'name': at.aum,
                'image': None,
                'type': None,
                'has_focus': at.disposition.has_focus,
                'title_location': at.disposition.title_location.name,
                'product_id': str(at.product_id),
                'sandbox_id': str(at.sandbox_id)
            }
            active_titles.append(title)

        status_json.update({'active_titles': active_titles})
        return status_json

    @property
    def media_status(self):
        if not self.usable or not self.console.media or not self.console.media.media_state:
            return None

        media_state = self.console.media.media_state

        # Ensure we are in the same app, otherwise this is useless
        if media_state.aum_id not in [t.aum for t in self.console.console_status.active_titles]:
            return None

        media_state_json = {
            'title_id': media_state.title_id,
            'aum_id': media_state.aum_id,
            'asset_id': media_state.asset_id,
            'media_type': media_state.media_type.name,
            'sound_level': media_state.sound_level.name,
            'enabled_commands': media_state.enabled_commands.value,
            'playback_status': media_state.playback_status.name,
            'rate': media_state.rate,
            'position': media_state.position,
            'media_start': media_state.media_start,
            'media_end': media_state.media_end,
            'min_seek': media_state.min_seek,
            'max_seek': media_state.max_seek,
            'metadata': None
        }

        metadata = {}
        for meta in media_state.metadata:
            metadata[meta.name] = meta.value

        media_state_json['metadata'] = metadata
        return media_state_json

    @property
    def status(self):
        data = self.console.to_dict()
        data.update({
            'connection_state': self.connection_state.name,
            'pairing_state': self.pairing_state.name,
            'device_status': self.device_status.name,
            'last_error': self.last_error,
            'authenticated_users_allowed': self.authenticated_users_allowed,
            'console_users_allowed': self.console_users_allowed,
            'anonymous_connection_allowed': self.anonymous_connection_allowed,
            'is_certificate_pending': self.is_certificate_pending
        })

        return data

    @property
    def stump_config(self):
        if self.usable:
            return self.console.stump.request_stump_configuration()

    @property
    def headend_info(self):
        if self.usable:
            return self.console.stump.request_headend_info()

    @property
    def livetv_info(self):
        if self.usable:
            return self.console.stump.request_live_tv_info()

    @property
    def tuner_lineups(self):
        if self.usable:
            return self.console.stump.request_tuner_lineups()

    @property
    def text_active(self):
        if self.usable:
            return self.console.text.got_active_session

    @property
    def nano_status(self):
        if not self.usable or 'nano' not in self.console.managers:
            return None

        nano = self.console.nano
        data = {
            'client_major_version': nano.client_major_version,
            'client_minor_version': nano.client_minor_version,
            'server_major_version': nano.server_major_version,
            'server_minor_version': nano.server_minor_version,
            'session_id': nano.session_id,
            'stream_can_be_enabled': nano.stream_can_be_enabled,
            'stream_enabled': nano.stream_enabled,
            'stream_state': nano.stream_state.name.lower(),
            'transmit_linkspeed': nano.transmit_linkspeed,
            'wireless': nano.wireless,
            'wireless_channel': nano.wireless_channel,
            'udp_port': nano.udp_port,
            'tcp_port': nano.tcp_port
        }
        return data

    def connect(self, userhash=None, xtoken=None):
        if not self.console:
            return enum.ConnectionState.Disconnected
        elif self.console.connected:
            return enum.ConnectionState.Connected
        elif not self.anonymous_connection_allowed and (not userhash or not xtoken):
            raise Exception('Requested anonymous connection is not allowed by console')

        state = self.console.connect(userhash=userhash,
                                     xsts_token=xtoken)

        if state == enum.ConnectionState.Connected:
            self.console.wait(0.5)
            self.console.stump.request_stump_configuration()

        return state

    def disconnect(self):
        self.console.disconnect()
        return True

    def power_off(self):
        self.console.power_off()
        return True

    def launch_title(self, app_id):
        return self.console.launch_title(app_id)

    def send_stump_key(self, device_id, button):
        result = self.console.send_stump_key(button, device_id)
        print(result)
        return True

    def send_media_command(self, command, seek_position=None):
        title_id = 0
        request_id = 0
        self.console.media_command(title_id, command, request_id, seek_position)
        return True

    def send_gamepad_button(self, btn):
        self.console.gamepad_input(btn)
        # Its important to clear button-press afterwards
        self.console.wait(0.1)
        self.console.gamepad_input(enum.GamePadButton.Clear)
        return True

    def send_text(self, text):
        if not self.text_active:
            return False

        self.console.send_systemtext_input(text)
        self.console.finish_text_input()
        return True

    def dvr_record(self, start_delta, end_delta):
        self.console.game_dvr_record(start_delta, end_delta)
        return True

    def nano_start(self):
        self.console.nano.start_stream()
        return True

    def nano_stop(self):
        self.console.nano.stop_stream()
        return True



from flask import current_app as app
from flask import request, render_template
from http import HTTPStatus
from xbox.webapi.authentication.manager import AuthenticationManager,\
    AuthenticationException, TwoFactorAuthRequired
from . import routes


@routes.route('/auth')
def authentication_overview():
    tokens = {
        'access_token': app.authentication_mgr.access_token,
        'refresh_token': app.authentication_mgr.refresh_token,
        'user_token': app.authentication_mgr.user_token,
        'xsts_token': app.authentication_mgr.xsts_token
    }

    data = {}
    for k, v in tokens.items():
        data.update({k: v.to_dict() if v else None})
    userinfo = app.authentication_mgr.userinfo.to_dict() if app.authentication_mgr.userinfo else None

    return app.success(tokens=data, userinfo=userinfo, authenticated=app.authentication_mgr.authenticated)


@routes.route('/auth/login')
def authentication_login():
    if app.authentication_mgr.authenticated:
        return render_template('auth_result.html',
                               title='Already signed in',
                               result='Already signed in',
                               message='You are already signed in, please logout first!',
                               link_path='/auth/logout',
                               link_title='Logout')
    else:
        return render_template('login.html')


@routes.route('/auth/login', methods=['POST'])
def authentication_login_post():
    is_webview = request.form.get('webview')
    email_address = request.form.get('email')
    password = request.form.get('password')

    if app.authentication_mgr.authenticated:
        return app.error('An account is already signed in.. please logout first', HTTPStatus.BAD_REQUEST)
    elif not email_address or not password:
        return app.error('No email or password parameter provided', HTTPStatus.BAD_REQUEST)

    app.authentication_mgr.email_address = email_address
    app.authentication_mgr.password = password

    try:
        app.authentication_mgr.authenticate()
        app.authentication_mgr.dump(app.token_file)
    except AuthenticationException as e:
        if is_webview:
            return render_template('auth_result.html',
                                   title='Login fail',
                                   result='Login failed',
                                   message='Error: {0}!'.format(str(e)),
                                   link_path='/auth/login',
                                   link_title='Try again')
        else:
            return app.error('Login failed! Error: {0}'.format(str(e)),
                             two_factor_required=False)

    except TwoFactorAuthRequired:
        if is_webview:
            return render_template('auth_result.html',
                                   title='Login fail',
                                   result='Login failed, 2FA required',
                                   message='Please click the following link to authenticate via OAUTH',
                                   link_path='/auth/oauth',
                                   link_title='Login via OAUTH')
        else:
            return app.error('Login failed, 2FA required!',
                             two_factor_required=True)

    except Exception as e:
        return app.error('Unhandled authentication error: {0}'.format(str(e)),
                         two_factor_required=False)

    if is_webview:
        return render_template('auth_result.html',
                               title='Login success',
                               result='Login succeeded',
                               message='Welcome {}!'.format(app.logged_in_gamertag),
                               link_path='/auth/logout',
                               link_title='Logout')
    else:
        return app.success(message='Login success', gamertag=app.logged_in_gamertag)


@routes.route('/auth/logout')
def authentication_logout():
    if app.authentication_mgr.authenticated:
        return render_template('logout.html', username=app.logged_in_gamertag)
    else:
        return render_template('auth_result.html',
                               title='Logout failed',
                               result='Logout failed',
                               message='You are currently not logged in',
                               link_path='/auth/login',
                               link_title='Login')


@routes.route('/auth/logout', methods=['POST'])
def authentication_logout_post():
    is_webview = request.form.get('webview')
    username = app.logged_in_gamertag
    app.reset_authentication()
    if is_webview:
        return render_template('auth_result.html',
                               title='Logout success',
                               result='Logout succeeded',
                               message='Goodbye {0}!'.format(username),
                               link_path='/auth/login',
                               link_title='Login')
    else:
        return app.success(message='Logout succeeded')


@routes.route('/auth/url')
def authentication_get_auth_url():
    return app.success(authorization_url=AuthenticationManager.generate_authorization_url())


@routes.route('/auth/oauth')
def authentication_oauth():
    if app.authentication_mgr.authenticated:
        return render_template('auth_result.html',
                               title='Already signed in',
                               result='Already signed in',
                               message='You are already signed in, please logout first!',
                               link_path='/auth/logout',
                               link_title='Logout')
    else:
        return render_template('login_oauth.html',
                               oauth_url=AuthenticationManager.generate_authorization_url())


@routes.route('/auth/oauth', methods=['POST'])
def authentication_oauth_post():
    is_webview = request.form.get('webview')
    app.reset_authentication()
    redirect_uri = request.form.get('redirect_uri')
    if not redirect_uri:
        return app.error('Please provide redirect_url', HTTPStatus.BAD_REQUEST)

    try:
        access, refresh = AuthenticationManager.parse_redirect_url(redirect_uri)
        app.authentication_mgr.access_token = access
        app.authentication_mgr.refresh_token = refresh
        app.authentication_mgr.authenticate(do_refresh=False)
        app.authentication_mgr.dump(app.token_file)
    except Exception as e:
        if is_webview:
            return render_template('auth_result.html',
                                   title='Login fail',
                                   result='Login failed',
                                   message='Error message: {0}'.format(str(e)),
                                   link_path='/auth/login',
                                   link_title='Try again')
        else:
            return app.error('Login failed, error: {0}'.format(str(e)))

    if is_webview:
        return render_template('auth_result.html',
                               title='Login success',
                               result='Login succeeded',
                               message='Welcome {}!'.format(app.logged_in_gamertag),
                               link_path='/auth/logout',
                               link_title='Logout')
    else:
        return app.success(message='Login success', gamertag=app.logged_in_gamertag)


@routes.route('/auth/refresh')
def authentication_refresh():
    try:
        app.authentication_mgr.authenticate(do_refresh=True)
    except Exception as e:
        return app.error(str(e))

    return app.success()


@routes.route('/auth/load')
def authentication_load_from_disk():
    try:
        app.authentication_mgr.load(app.token_file)
    except FileNotFoundError as e:
        return app.error('Failed to load tokens from \'{0}\'. Error: {1}'.format(e.filename, e.strerror), HTTPStatus.NOT_FOUND)

    return app.success()


@routes.route('/auth/store')
def authentication_store_on_disk():
    if not app.authentication_mgr.authenticated:
        return app.error('Sorry, no valid authentication for saving was found', HTTPStatus.BAD_REQUEST)

    try:
        app.authentication_mgr.dump(app.token_file)
    except Exception as e:
        return app.error('Failed to save tokens to \'{0}\'. Error: {1}'.format(app.token_file, str(e)))

    return app.success()



from flask import current_app as app
from flask import request, jsonify
from ..decorators import require_authentication
from . import routes


@routes.route('/web/title/<title_id>')
@require_authentication
def download_title_info(client, title_id):
    try:
        resp = client.titlehub.get_title_info(title_id, 'image').json()
        return jsonify(resp['titles'][0])
    except KeyError:
        return app.error('Cannot find titles-node json response')
    except IndexError:
        return app.error('No info for requested title not found')
    except Exception as e:
        return app.error('Download of titleinfo failed, error: {0}'.format(e))


@routes.route('/web/titlehistory')
@require_authentication
def download_title_history(client):
    try:
        max_items = request.args.get('max_items') or 5
        resp = client.titlehub.get_title_history(app.xbl_client.xuid, max_items=max_items).json()
        return jsonify(resp)
    except Exception as e:
        return app.error('Download of titlehistory failed, error: {0}'.format(e))


@routes.route('/web/pins')
@require_authentication
def download_pins(client):
    try:
        resp = client.lists.get_items(app.xbl_client.xuid, {}).json()
        return jsonify(resp)
    except Exception as e:
        return app.error('Download of pins failed, error: {0}'.format(e))



"""
Stump enumerations
"""

from uuid import UUID
from enum import Enum


class Message(Enum):
    """
    Message types
    """
    ERROR = "Error"
    ENSURE_STREAMING_STARTED = "EnsureStreamingStarted"
    CONFIGURATION = "GetConfiguration"
    HEADEND_INFO = "GetHeadendInfo"
    LIVETV_INFO = "GetLiveTVInfo"
    PROGRAMM_INFO = "GetProgrammInfo"
    RECENT_CHANNELS = "GetRecentChannels"
    TUNER_LINEUPS = "GetTunerLineups"
    APPCHANNEL_DATA = "GetAppChannelData"
    APPCHANNEL_LINEUPS = "GetAppChannelLineups"
    APPCHANNEL_PROGRAM_DATA = "GetAppChannelProgramData"
    SEND_KEY = "SendKey"
    SET_CHANNEL = "SetChannel"


class Notification(Enum):
    """
    Notification types
    """
    STREAMING_ERROR = "StreamingError"
    CHANNEL_CHANGED = "ChannelChanged"
    CHANNELTYPE_CHANGED = "ChannelTypeChanged"
    CONFIGURATION_CHANGED = "ConfigurationChanged"
    DEVICE_UI_CHANGED = "DeviceUIChanged"
    HEADEND_CHANGED = "HeadendChanged"
    VIDEOFORMAT_CHANGED = "VideoFormatChanged"
    PROGRAM_CHANGED = "ProgrammChanged"
    TUNERSTATE_CHANGED = "TunerStateChanged"


class Source(Enum):
    """
    Streamingsources
    """
    HDMI = "hdmi"
    TUNER = "tuner"


class DeviceType(Enum):
    """
    Devicetypes
    """
    TV = "tv"
    TUNER = "tuner"
    SET_TOP_BOX = "stb"
    AV_RECEIVER = "avr"


class SourceHttpQuery(Enum):
    """
    Source strings used in HTTP query
    """
    HDMI = "hdmi-in"
    TUNER = "zurich"


class Input(object):
    HDMI = UUID("BA5EBA11-DEA1-4BAD-BA11-FEDDEADFAB1E")


class Quality(Enum):
    """
    Quality values
    """
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"
    BEST = "best"


class FilterType(Enum):
    """
    Channel-filter types
    """
    ALL = "ALL"    # No filter / Show all
    HDSD = "HDSD"  # Dont show double SD-channels
    HD = "HD"      # Only show HD Channels



from http import HTTPStatus


def test_versions(rest_client):
    resp = rest_client.test_client().get('/versions')

    assert resp.status_code == HTTPStatus.OK
    assert resp.json['success'] is True

    assert 'versions' in resp.json
    assert len(resp.json['versions']) > 1



"""
Smartglass message un-/packing

Also handles de/encryption internally, this means:
You feed plaintext data (wrapped in :class:`.XStruct`) and you get plaintext data

**Note on encryption**
Depending on the packet-type, acquiring the `Initialization Vector` and
encrypting the `protected payload` happens differently:

* ConnectRequest: The `IV` is randomly chosen from calculated
  Elliptic Curve (which happens in :class:`Crypto`) and is delivered
  inside the `unprotected payload` section of the ConnectRequest message.
* Messages: The `IV` is generated by encrypting the first 16 bytes of
  the `unencrypted` header.

**Note on padding**
If message has `protected payload` it might need padding according
to PKCS#7 (e.g. padding is in whole bytes, the value of each added byte is
the number of bytes that are added, i.e. N bytes, each of value N are
added. thx wikipedia).
"""
from construct import Int16ub
from xbox.sg.enum import PacketType
from xbox.sg.crypto import PKCS7Padding
from xbox.sg.packet import simple, message
from xbox.sg.utils.struct import flatten


class PackerError(Exception):
    """
    Custom exceptions for usage in packer module
    """
    pass


def unpack(buf, crypto=None):
    """
    Unpacks messages from Smartglass CoreProtocol.

    For messages that require decryption, a Crypto instance needs to be passed
    as well.

    The decryption happens in :class:`CryptoTunnel`.

    Args:
        buf (bytes): A byte string to be deserialized into a message.
        crypto (Crypto): Instance of :class:`Crypto`.

    Raises:
        PackerError: On various errors, instance of :class:`PackerError`.

    Returns:
        Container: The deserialized message, instance of :class:`Container`.
    """
    msg_struct = None
    pkt_type = PacketType(Int16ub.parse(buf[:2]))

    if pkt_type not in PacketType:
        raise PackerError("Invalid packet type")

    if pkt_type in simple.pkt_types:
        msg_struct = simple.struct
    elif pkt_type == PacketType.Message:
        msg_struct = message.struct

    return msg_struct.parse(buf, _crypto=crypto)


def pack(msg, crypto=None):
    """
    Packs messages for Smartglass CoreProtocol.

    For messages that require encryption, a Crypto instance needs to be passed
    as well.

    Args:
        msg (XStructObj): A serializable message, instance of
                          :class:`XStructObj`.
        crypto (Crypto): Instance of :class:`Crypto`.

    Returns:
        bytes: The serialized bytes.
    """
    container = flatten(msg.container)
    packed_unprotected = b''
    packed_protected = b''

    if container.get('unprotected_payload', None):
        packed_unprotected = msg.subcon.unprotected_payload.build(
            container.unprotected_payload, **container
        )

    if container.get('protected_payload', None):
        packed_protected = msg.subcon.protected_payload.build(
            container.protected_payload, **container
        )

    container.header.unprotected_payload_length = len(packed_unprotected)
    container.header.protected_payload_length = len(packed_protected)
    packed_header = msg.subcon.header.build(
        container.header, **container
    )

    if container.get('protected_payload', None) and packed_protected:
        connect_types = [PacketType.ConnectRequest, PacketType.ConnectResponse]
        if container.header.pkt_type in connect_types:
            iv = container.unprotected_payload.iv
        elif container.header.pkt_type == PacketType.Message:
            iv = crypto.generate_iv(packed_header[:16])
        else:
            raise PackerError("Incompatible packet type for encryption")

        if PKCS7Padding.size(len(packed_protected), 16) > 0:
            packed_protected = PKCS7Padding.pad(packed_protected, 16)

        packed_protected = crypto.encrypt(iv, packed_protected)
        buffer = packed_header + packed_unprotected + packed_protected
        return buffer + crypto.hash(buffer)
    else:
        return packed_header + packed_unprotected


def payload_length(msg):
    """
    Calculates the packed length in bytes of the given message.

    Args:
        msg (XStructObj): A serializable message, instance of
                          :class:`XStructObj`.

    Returns:
        int: The packed message length in bytes.
    """
    container = flatten(msg.container)
    packed_unprotected = b''
    packed_protected = b''

    if container.get('unprotected_payload', None):
        packed_unprotected = msg.subcon.unprotected_payload.build(
            container.unprotected_payload, **container
        )

    if container.get('protected_payload', None):
        packed_protected = msg.subcon.protected_payload.build(
            container.protected_payload, **container
        )

    return len(packed_unprotected + packed_protected)



"""
Console

Console class is a wrapper around the protocol, can be used to
`discover`, `poweron` and `connect`.
Also stores various device status informations.

It can be either created manually or instantiated via `DiscoveryResponse`
message. However, calling static method `discover` does all that for you
automatically.

Example:
    Discovery and connecting::

        import sys
        from xbox.sg.console import Console
        from xbox.sg.enum import ConnectionState

        discovered = Console.discover(timeout=1)
        if len(discovered):
            console = discovered[0]
            console.connect()
            if console.connection_state != ConnectionState.Connected:
                print("Connection failed")
                sys.exit(1)
            console.wait(1)

            try:
                console.protocol.serve_forever()
            except KeyboardInterrupt:
                pass
        else:
            print("No consoles discovered")
            sys.exit(1)
"""

import gevent
from uuid import UUID

from xbox.sg.crypto import Crypto
from xbox.sg.manager import Manager
from xbox.sg.enum import PairedIdentityState, DeviceStatus, ConnectionState,\
    MessageType, PrimaryDeviceFlag, ActiveTitleLocation
from xbox.sg.protocol import CoreProtocol, ProtocolError
from xbox.sg.utils.events import Event


class Console(object):
    __protocol__ = CoreProtocol()

    def __init__(self, address, name, uuid, liveid, flags=PrimaryDeviceFlag.Null, last_error=0, public_key=None):
        """
        Initialize an instance of Console

        Args:
            address (str): IP address of console.
            flags (:class:`PrimaryDeviceFlag`): Primary device flags
            name (str): Name of console.
            uuid (:obj:`UUID`): UUID of console.
            liveid (str): Live ID of console.
            public_key (:obj:`ec.EllipticCurvePrivateKey`):
                Console's Public Key.
        """
        self.address = address
        self.name = name
        self.uuid = uuid
        self.liveid = liveid
        self.flags = flags
        self.last_error = last_error
        self._public_key = None
        self._crypto = None

        if public_key:
            # This sets up the crypto context
            self.public_key = public_key

        self._device_status = DeviceStatus.Unavailable
        self._connection_state = ConnectionState.Disconnected
        self._pairing_state = PairedIdentityState.NotPaired
        self._console_status = None
        self._active_surface = None

        self.on_device_status = Event()
        self.on_connection_state = Event()
        self.on_pairing_state = Event()
        self.on_console_status = Event()
        self.on_active_surface = Event()
        self.on_timeout = Event()

        self.managers = {}
        self._functions = {}

        self.power_on = self._power_on  # Dirty hack

        self._init_protocol()

    def __repr__(self):
        return '<Console addr={} name={} uuid={} liveid={} flags={} last_error={}>'.format(
            self.address, self.name, self.uuid, self.liveid, self.flags, self.last_error
        )

    def __enter__(self):
        return self

    def __exit__(self):
        self.disconnect()

    @staticmethod
    def wait(seconds):
        """
        Wrapper around `gevent.sleep`

        Args:
            seconds (int): Seconds to wait.

        Returns: None
        """
        gevent.sleep(seconds)

    def _init_protocol(self):
        self.protocol = CoreProtocol(self.address, self._crypto)
        # Proxy protocol event into console
        self.protocol.on_timeout += self._on_timeout
        self.protocol.on_message += self._on_message

    @classmethod
    def __start_cls_protocol(cls):
        """
        Start the protocol internally, if not running

        Returns: None
        """
        if not cls.__protocol__.started:
            cls.__protocol__.start()

    @classmethod
    def from_message(cls, address, msg):
        """
        Initialize the class with a `DiscoveryResponse`.

        Args:
            address (str): IP address of the console
            msg (:obj:`.StructObj`): Discovery Response struct

        Returns: :class:`Console`.

        """
        payload = msg.unprotected_payload
        console = cls(
            address, payload.name, payload.uuid, payload.cert.liveid,
            payload.flags, payload.last_error, payload.cert.pubkey
        )
        console.device_status = DeviceStatus.Available
        return console

    @classmethod
    def from_dict(cls, d):
        return cls(d['address'], d['name'], UUID(d['uuid']), d['liveid'])

    def to_dict(self):
        return dict(address=self.address, name=self.name, uuid=str(self.uuid), liveid=self.liveid)

    def add_manager(self, manager, *args, **kwargs):
        """
        Add a manager to the console instance.

        This will inherit all public methods of the manager class.

        Args:
            manager (:class:`.Manager`):
            *args: Arguments
            **kwargs: KwArguments

        Returns:
            None
        """
        if not issubclass(manager, Manager):
            raise ValueError("Manager needs to subclass {}.{}".format(
                Manager.__module__, Manager.__name__)
            )

        namespace = getattr(manager, '__namespace__', manager.__name__.lower())
        manager_inst = manager(self, *args, **kwargs)
        self.managers[namespace] = manager_inst

        for item in dir(manager):
            if item.startswith('_'):
                continue

            if item in self.__dict__:
                raise ValueError("Attribute already exists: %s" % item)

            self._functions[item] = getattr(manager_inst, item)

    def __getattr__(self, k):
        if k in self.managers:
            return self.managers[k]
        elif k in self._functions:
            return self._functions[k]
        return object.__getattribute__(self, k)

    @classmethod
    def discover(cls, *args, **kwargs):
        """
        Discover consoles on the network.

        Args:
            *args:
            **kwargs:

        Returns:
            list: List of discovered consoles.

        """
        cls.__start_cls_protocol()
        discovered = cls.__protocol__.discover(*args, **kwargs)
        return [cls.from_message(a, m) for a, m in discovered.items()]

    @classmethod
    def discovered(cls):
        """
        Get list of already discovered consoles.

        Returns:
            list: List of discovered consoles.

        """
        discovered = cls.__protocol__.discovered
        return [cls.from_message(a, m) for a, m in discovered.items()]

    @classmethod
    def power_on(cls, liveid, addr=None, tries=2):
        """
        Power On console with given Live ID.

        Optionally the IP address of the console can be supplied,
        this is useful if the console is stubborn and does not react
        to broadcast / multicast packets (due to routing issues).

        Args:
            liveid (str): Live ID of console.
            addr (str): IP address of console.
            tries (int): Poweron attempts, default: 2.

        Returns: None

        """
        cls.__start_cls_protocol()
        cls.__protocol__.power_on(liveid, addr, tries)

    def _power_on(self, tries=2):
        Console.power_on(self.liveid, self.address, tries)

    def connect(self, userhash=None, xsts_token=None):
        """
        Connect to the console

        If the connection fails, error will be stored in
        `self.connection_state`

        Raises:
            ConnectionException: If no authentication data is supplied and
                console disallows anonymous connection.

        Returns:
            :class:`ConnectionState`: Connection state
        """
        if self.connected:
            raise ConnectionError("Already connected")

        auth_data_available = bool(userhash and xsts_token)
        if not self.anonymous_connection_allowed and not auth_data_available:
            raise ConnectionError("Anonymous connection not allowed, please"
                                  " supply userhash and auth-token")

        if not self.protocol.started:
            self.protocol.start()

        self.pairing_state = PairedIdentityState.NotPaired
        self.connection_state = ConnectionState.Connecting

        try:
            self.pairing_state = self.protocol.connect(userhash=userhash,
                                                       xsts_token=xsts_token)
        except ProtocolError as e:
            self.connection_state = ConnectionState.Error
            raise e

        self.connection_state = ConnectionState.Connected
        return self.connection_state

    def launch_title(self, uri, location=ActiveTitleLocation.Full):
        """
        Launch a title by URI

        Args:
            uri (str): Launch uri
            location (:class:`ActiveTitleLocation`): Target title location

        Returns:
            int: Member of :class:`AckStatus`
        """
        return self.protocol.launch_title(uri, location)

    def game_dvr_record(self, start_delta, end_delta):
        """
        Start Game DVR recording

        Args:
            start_delta (int): Start time
            end_delta (int): End time

        Returns:
            int: Member of :class:`AckStatus`
        """
        return self.protocol.game_dvr_record(start_delta, end_delta)

    def disconnect(self):
        """
        Disconnect from console.

        This will reset connection-, pairing-state,
        ActiveSurface and ConsoleStatus.

        Returns: None
        """
        if self.connection_state == ConnectionState.Connected:
            self.connection_state = ConnectionState.Disconnecting
            self.protocol.stop()  # Stop also sends SG disconnect message
            self._reset_state()

    def power_off(self):
        """
        Power off the console.

        No need to disconnect after.

        Returns: None
        """
        self.protocol.power_off(self.liveid)
        self._reset_state()
        self.device_status = DeviceStatus.Unavailable

    def _on_message(self, msg, channel):
        """
        Internal handler for console specific messages aka.
        `PairedIdentityStateChange`, `ConsoleStatus` and
        `ActiveSurfaceChange`.

        Args:
            msg (:obj:`XStruct`): Message data
            channel (:class:`ServiceChannel`): Service channel

        Returns:
            None
        """
        msg_type = msg.header.flags.msg_type

        if msg_type == MessageType.PairedIdentityStateChanged:
            self.pairing_state = msg.protected_payload.state

        elif msg_type == MessageType.ConsoleStatus:
            self.console_status = msg.protected_payload

        elif msg_type == MessageType.ActiveSurfaceChange:
            self.active_surface = msg.protected_payload

    def _on_timeout(self):
        """
        Internal handler for console connection timeout.

        Returns: None
        """
        self._reset_state()
        self.device_status = DeviceStatus.Unavailable
        self.on_timeout()

    def _reset_state(self):
        """
        Internal handler to reset the inital state of the console instance.

        Returns: None
        """
        if self.protocol and self.protocol.started:
            self.protocol.stop()
        self._init_protocol()

        self.connection_state = ConnectionState.Disconnected
        self.pairing_state = PairedIdentityState.NotPaired
        self.active_surface = None
        self.console_status = None

    @property
    def public_key(self):
        """
        Console's public key.

        Returns:
            :obj:`ec.EllipticCurvePrivateKey`: Foreign public key
        """
        return self._public_key

    @public_key.setter
    def public_key(self, key):
        if isinstance(key, bytes):
            self._crypto = Crypto.from_bytes(key)
        else:
            self._crypto = Crypto(key)

        self._public_key = self._crypto.foreign_pubkey

    @property
    def device_status(self):
        """
        Current availability status

        Returns:
            :obj:`DeviceStatus`
        """
        return self._device_status

    @device_status.setter
    def device_status(self, status):
        self._device_status = status
        self.on_device_status(status)

    @property
    def connection_state(self):
        """
        Current connection state

        Returns:
            :obj:`ConnectionState`
        """
        return self._connection_state

    @connection_state.setter
    def connection_state(self, state):
        self._connection_state = state
        self.on_connection_state(state)

    @property
    def pairing_state(self):
        """
        Current pairing state

        Returns:
            :obj:`PairedIdentityState`
        """
        return self._pairing_state

    @pairing_state.setter
    def pairing_state(self, state):
        self._pairing_state = state
        self.on_pairing_state(state)

    @property
    def console_status(self):
        """
        Console status aka. kernel version, active titles etc.

        Returns:
            :obj:`XStruct`
        """
        return self._console_status

    @console_status.setter
    def console_status(self, status):
        self._console_status = status
        self.on_console_status(status)

    @property
    def active_surface(self):
        """
        Currently active surface

        Returns:
            :obj:`XStruct`
        """
        return self._active_surface

    @active_surface.setter
    def active_surface(self, active_surface):
        self._active_surface = active_surface
        self.on_active_surface(active_surface)

    @property
    def available(self):
        """
        Check whether console is available aka. discoverable

        Returns:
            bool: `True` if console is available, `False` otherwise
        """
        return self.device_status == DeviceStatus.Available

    @property
    def paired(self):
        """
        Check whether client is paired to console

        Returns:
            bool: `True` if console is paired, `False` otherwise
        """
        return self.pairing_state == PairedIdentityState.Paired

    @property
    def connected(self):
        """
        Check whether client is successfully connected to console

        Returns:
            bool: `True` if connected, `False` otherwise
        """
        return self.connection_state == ConnectionState.Connected

    @property
    def authenticated_users_allowed(self):
        """
        Check whether authenticated users are allowed to connect

        Returns:
            bool: `True` if authenticated users are allowed, `False` otherwise
        """
        return bool(self.flags & PrimaryDeviceFlag.AllowAuthenticatedUsers)

    @property
    def console_users_allowed(self):
        """
        Check whether console users are allowed to connect

        Returns:
            bool: `True` if console users are allowed, `False` otherwise
        """
        return bool(self.flags & PrimaryDeviceFlag.AllowConsoleUsers)

    @property
    def anonymous_connection_allowed(self):
        """
        Check whether anonymous connection is allowed

        Returns:
            bool: `True` if anonymous connection is allowed, `False` otherwise
        """
        return bool(self.flags & PrimaryDeviceFlag.AllowAnonymousUsers)

    @property
    def is_certificate_pending(self):
        """
        Check whether certificate is pending

        Returns:
            bool: `True` if certificate is pending, `False` otherwise
        """
        return bool(self.flags & PrimaryDeviceFlag.CertificatePending)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup


setup(
    name="xbox-smartglass-core",
    version="1.2.2",
    author="OpenXbox",
    author_email="noreply@openxbox.org",
    description="A library to interact with the Xbox One gaming console via the SmartGlass protocol.",
    long_description=open('README.rst').read() + '\n\n' + open('HISTORY.rst').read(),
    long_description_content_type="text/x-rst",
    license="GPL",
    keywords="xbox one smartglass auxiliary fallout title stump tv streaming livetv rest api",
    url="https://github.com/OpenXbox/xbox-smartglass-core-python",
    python_requires=">=3.6",
    packages=[
        'xbox.sg',
        'xbox.sg.utils',
        'xbox.sg.packet',
        'xbox.auxiliary',
        'xbox.stump',
        'xbox.rest',
        'xbox.rest.routes',
        'xbox.scripts',
        'xbox.handlers'
    ],
    namespace_packages=['xbox'],
    zip_safe=False,
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8"
    ],
    install_requires=[
        'xbox-webapi>=1.1.8',
        'construct==2.10.56',
        'cryptography==2.8',
        'gevent==1.5a3',
        'dpkt',
        'marshmallow-objects',
        'Flask'
    ],
    setup_requires=['pytest-runner'],
    tests_require=[
        'pytest',
        'pytest-console-scripts',
        'flake8',
        'tox'
    ],
    extras_require={
        'dev': [
            'bump2version',
            'watchdog',
            'Sphinx',
            'wheel',
            'twine',
            'coverage',
            'pytest',
            'pytest-runner',
            'pytest-console-scripts'
        ]
    },
    test_suite="tests",
    entry_points={
        'console_scripts': [
            'xbox-cli=xbox.scripts.main_cli:main',
            'xbox-discover=xbox.scripts.main_cli:main_discover',
            'xbox-poweron=xbox.scripts.main_cli:main_poweron',
            'xbox-poweroff=xbox.scripts.main_cli:main_poweroff',
            'xbox-repl=xbox.scripts.main_cli:main_repl',
            'xbox-replserver=xbox.scripts.main_cli:main_replserver',
            'xbox-textinput=xbox.scripts.main_cli:main_textinput',
            'xbox-gamepadinput=xbox.scripts.main_cli:main_gamepadinput',
            'xbox-tui=xbox.scripts.main_cli:main_tui',
            'xbox-fo4-relay=xbox.scripts.main_cli:main_falloutrelay',
            'xbox-pcap=xbox.scripts.pcap:main',
            'xbox-recrypt=xbox.scripts.recrypt:main',
            'xbox-rest-server=xbox.scripts.rest_server:main'
        ]
    }
)



from xbox.sg.crypto import Padding, PKCS7Padding, ANSIX923Padding


def test_calculate_padding():
    align_16 = Padding.size(12, alignment=16)
    align_12 = Padding.size(12, alignment=12)
    align_10 = Padding.size(12, alignment=10)

    assert align_16 == 4
    assert align_12 == 0
    assert align_10 == 8


def test_remove_padding():
    payload = 8 * b'\x88' + b'\x00\x00\x00\x04'
    unpadded = Padding.remove(payload)

    assert len(unpadded) == 8
    assert unpadded == 8 * b'\x88'


def test_x923_add_padding():
    payload = 7 * b'\x69'
    padded_12 = ANSIX923Padding.pad(payload, alignment=12)
    padded_7 = ANSIX923Padding.pad(payload, alignment=7)
    padded_3 = ANSIX923Padding.pad(payload, alignment=3)

    assert len(padded_12) == 12
    assert len(padded_7) == 7
    assert len(padded_3) == 9
    assert padded_12 == payload + b'\x00\x00\x00\x00\x05'
    assert padded_7 == payload
    assert padded_3 == payload + b'\x00\x02'


def test_pkcs7_add_padding():
    payload = 7 * b'\x69'
    padded_12 = PKCS7Padding.pad(payload, alignment=12)
    padded_7 = PKCS7Padding.pad(payload, alignment=7)
    padded_3 = PKCS7Padding.pad(payload, alignment=3)

    assert len(padded_12) == 12
    assert len(padded_7) == 7
    assert len(padded_3) == 9
    assert padded_12 == payload + b'\x05\x05\x05\x05\x05'
    assert padded_7 == payload
    assert padded_3 == payload + b'\x02\x02'



import logging

import gevent
from gevent.server import StreamServer

from xbox.sg.crypto import PKCS7Padding
from xbox.sg.utils.events import Event
from xbox.auxiliary import packer
from xbox.auxiliary.packet import aux_header_struct, AUX_PACKET_MAGIC
from xbox.auxiliary.crypto import AuxiliaryStreamCrypto

log = logging.getLogger(__name__)


class AuxiliaryPackerException(Exception):
    pass


class ConsoleConnection(object):
    BUFFER_SIZE = 2048

    def __init__(self, address, port, crypto):
        self.host = (address, port)
        self.crypto = crypto

        self._socket = None
        self._recv_thread = None

        self.on_message = Event()

    def start(self):
        self._socket = gevent.socket.create_connection(self.host)
        self._recv_thread = gevent.spawn(self._recv)

    def stop(self):
        self._socket.close()
        gevent.kill(self._recv_thread)

    def handle(self, data):
        try:
            msg = packer.unpack(data, self.crypto)
            # Fire event
            self.on_message(msg)
        except Exception as e:
            log.exception("Exception while handling Console Aux data, error: {}".format(e))

    def _recv(self):
        while True:
            gevent.socket.wait_read(self._socket.fileno())
            data = self._socket.recv(4)
            header = aux_header_struct.parse(data)

            if header.magic != AUX_PACKET_MAGIC:
                raise Exception('Invalid packet magic received from console')

            payload_sz = header.payload_size + PKCS7Padding.size(header.payload_size, 16)
            remaining_payload_bytes = payload_sz

            while remaining_payload_bytes > 0:
                gevent.socket.wait_read(self._socket.fileno())
                tmp = self._socket.recv(remaining_payload_bytes)
                remaining_payload_bytes -= len(tmp)
                data += tmp

            data += self._socket.recv(32)

            self.handle(data)

    def send(self, msg):
        packets = packer.pack(msg, self.crypto)

        if not packets:
            raise Exception('No data')

        for packet in packets:
            self._socket.send(packet)


class AuxiliaryRelayService(object):
    def __init__(self, connection_info, listen_port):
        if len(connection_info.endpoints) > 1:
            raise Exception('Auxiliary Stream advertises more than one endpoint!')

        self.crypto = AuxiliaryStreamCrypto.from_connection_info(connection_info)
        self.target_ip = connection_info.endpoints[0].ip
        self.target_port = connection_info.endpoints[0].port

        self.console_connection = ConsoleConnection(self.target_ip, self.target_port, self.crypto)
        self.server = StreamServer(('0.0.0.0', listen_port), self._handle_client)
        self.client_socket = None

    def run(self):
        self.server.start()

    def _handle_client(self, socket, addr):
        self.client_socket = socket

        self.console_connection.on_message += self._handle_console_data
        self.console_connection.start()

        while True:
            gevent.socket.wait_read(self.client_socket.fileno())
            data = self.client_socket.recv(2048)
            if not data:
                self.client_socket.close()
                self.client_socket = None
                log.warning('Client Socket closed!\n')
                break
            else:
                self._handle_client_data(data)

    def _handle_console_data(self, data):
        # Data from console gets decrypted and forwarded to aux client
        if self.client_socket:
            self.client_socket.send(data)

    def _handle_client_data(self, data):
        # Data from aux client gets encrypted and sent to console
        self.console_connection.send(data)



# flake8: noqa
"""
Construct containers for message header and payloads
"""
from construct import *
from xbox.sg import enum
from xbox.sg.enum import PacketType, MessageType
from xbox.sg.utils.struct import XStruct
from xbox.sg.utils.adapters import CryptoTunnel, UUIDAdapter, JsonAdapter, XSwitch, XEnum, SGString, PrefixedBytes


header = XStruct(
    'pkt_type' / Default(XEnum(Int16ub, PacketType), PacketType.Message),
    'protected_payload_length' / Default(Int16ub, 0),
    'sequence_number' / Int32ub,
    'target_participant_id' / Int32ub,
    'source_participant_id' / Int32ub,
    'flags' / BitStruct(
        'version' / Default(BitsInteger(2), 2),
        'need_ack' / Flag,
        'is_fragment' / Flag,
        'msg_type' / XEnum(BitsInteger(12), MessageType)
    ),
    'channel_id' / Int64ub
)


fragment = XStruct(
    'sequence_begin' / Int32ub,
    'sequence_end' / Int32ub,
    'data' / PrefixedBytes(Int16ub)
)


acknowledge = XStruct(
    'low_watermark' / Int32ub,
    'processed_list' / PrefixedArray(Int32ub, Int32ub),
    'rejected_list' / PrefixedArray(Int32ub, Int32ub)
)


json = XStruct(
    'text' / JsonAdapter(SGString())
)


local_join = XStruct(
    'device_type' / XEnum(Int16ub, enum.ClientType),
    'native_width' / Int16ub,
    'native_height' / Int16ub,
    'dpi_x' / Int16ub,
    'dpi_y' / Int16ub,
    'device_capabilities' / XEnum(Int64ub, enum.DeviceCapabilities),
    'client_version' / Int32ub,
    'os_major_version' / Int32ub,
    'os_minor_version' / Int32ub,
    'display_name' / SGString()
)


auxiliary_stream = XStruct(
    'connection_info_flag' / Int8ub,
    'connection_info' / If(this.connection_info_flag == 1, Struct(
        'crypto_key' / PrefixedBytes(Int16ub),
        'server_iv' / PrefixedBytes(Int16ub),
        'client_iv' / PrefixedBytes(Int16ub),
        'sign_hash' / PrefixedBytes(Int16ub),
        'endpoints' / PrefixedArray(Int16ub, Struct(
            'ip' / SGString(),
            'port' / SGString()
        ))
    ))
)


power_off = XStruct(
    'liveid' / SGString()
)


game_dvr_record = XStruct(
    'start_time_delta' / Int32sb,
    'end_time_delta' / Int32sb
)


unsnap = XStruct(
    'unk' / Bytes(1)
)


gamepad = XStruct(
    'timestamp' / Int64ub,
    'buttons' / XEnum(Int16ub, enum.GamePadButton),
    'left_trigger' / Float32b,
    'right_trigger' / Float32b,
    'left_thumbstick_x' / Float32b,
    'left_thumbstick_y' / Float32b,
    'right_thumbstick_x' / Float32b,
    'right_thumbstick_y' / Float32b
)


paired_identity_state_changed = XStruct(
    'state' / XEnum(Int16ub, enum.PairedIdentityState)
)


media_state = XStruct(
    'title_id' / Int32ub,
    'aum_id' / SGString(),
    'asset_id' / SGString(),
    'media_type' / XEnum(Int16ub, enum.MediaType),
    'sound_level' / XEnum(Int16ub, enum.SoundLevel),
    'enabled_commands' / XEnum(Int32ub, enum.MediaControlCommand),
    'playback_status' / XEnum(Int16ub, enum.MediaPlaybackStatus),
    'rate' / Float32b,
    'position' / Int64ub,
    'media_start' / Int64ub,
    'media_end' / Int64ub,
    'min_seek' / Int64ub,
    'max_seek' / Int64ub,
    'metadata' / PrefixedArray(Int16ub, Struct(
        'name' / SGString(),
        'value' / SGString()
    ))
)


media_controller_removed = XStruct(
    'title_id' / Int32ub
)


media_command_result = XStruct(
    'request_id' / Int64ub,
    'result' / Int32ub
)


media_command = XStruct(
    'request_id' / Int64ub,
    'title_id' / Int32ub,
    'command' / XEnum(Int32ub, enum.MediaControlCommand),
    'seek_position' / If(this.command == enum.MediaControlCommand.Seek, Int64ub)
)


orientation = XStruct(
    'timestamp' / Int64ub,
    'rotation_matrix_value' / Float32b,
    'w' / Float32b,
    'x' / Float32b,
    'y' / Float32b,
    'z' / Float32b
)


compass = XStruct(
    'timestamp' / Int64ub,
    'magnetic_north' / Float32b,
    'true_north' / Float32b
)


inclinometer = XStruct(
    'timestamp' / Int64ub,
    'pitch' / Float32b,
    'roll' / Float32b,
    'yaw' / Float32b
)


gyrometer = XStruct(
    'timestamp' / Int64ub,
    'angular_velocity_x' / Float32b,
    'angular_velocity_y' / Float32b,
    'angular_velocity_z' / Float32b
)


accelerometer = XStruct(
    'timestamp' / Int64ub,
    'acceleration_x' / Float32b,
    'acceleration_y' / Float32b,
    'acceleration_z' / Float32b
)


_touchpoint = XStruct(
    'touchpoint_id' / Int32ub,
    'touchpoint_action' / XEnum(Int16ub, enum.TouchAction),
    'touchpoint_x' / Int32ub,
    'touchpoint_y' / Int32ub
)


touch = XStruct(
    'touch_msg_timestamp' / Int32ub,
    'touchpoints' / PrefixedArray(Int16ub, _touchpoint)
)


disconnect = XStruct(
    'reason' / XEnum(Int32ub, enum.DisconnectReason),
    'error_code' / Int32ub
)


stop_channel = XStruct(
    'target_channel_id' / Int64ub
)


start_channel_request = XStruct(
    'channel_request_id' / Int32ub,
    'title_id' / Int32ub,
    'service' / UUIDAdapter(),
    'activity_id' / Int32ub
)


start_channel_response = XStruct(
    'channel_request_id' / Int32ub,
    'target_channel_id' / Int64ub,
    'result' / XEnum(Int32ub, enum.SGResultCode)
)


title_launch = XStruct(
    'location' / XEnum(Int16ub, enum.ActiveTitleLocation),
    'uri' / SGString()
)


system_text_done = XStruct(
    'text_session_id' / Int32ub,
    'text_version' / Int32ub,
    'flags' / Int32ub,
    'result' / XEnum(Int32ub, enum.TextResult)
)


system_text_acknowledge = XStruct(
    'text_session_id' / Int32ub,
    'text_version_ack' / Int32ub
)

_system_text_input_delta = XStruct(
    'offset' / Int32ub,
    'delete_count' / Int32ub,
    'insert_content' / SGString()
)

system_text_input = XStruct(
    'text_session_id' / Int32ub,
    'base_version' / Int32ub,
    'submitted_version' / Int32ub,
    'total_text_byte_len' / Int32ub,
    'selection_start' / Int32sb,
    'selection_length' / Int32sb,
    'flags' / Int16ub,
    'text_chunk_byte_start' / Int32ub,
    'text_chunk' / SGString(),
    'delta' / Optional(PrefixedArray(Int16ub, _system_text_input_delta))
)


title_text_selection = XStruct(
    'text_session_id' / Int64ub,
    'text_buffer_version' / Int32ub,
    'start' / Int32ub,
    'length' / Int32ub
)


title_text_input = XStruct(
    'text_session_id' / Int64ub,
    'text_buffer_version' / Int32ub,
    'result' / XEnum(Int16ub, enum.TextResult),
    'text' / SGString()
)


text_configuration = XStruct(
    'text_session_id' / Int64ub,
    'text_buffer_version' / Int32ub,
    'text_options' / XEnum(Int32ub, enum.TextOption),
    'input_scope' / XEnum(Int32ub, enum.TextInputScope),
    'max_text_length' / Int32ub,
    'locale' / SGString(),
    'prompt' / SGString()
)


_active_title = XStruct(
    'title_id' / Int32ub,
    'disposition' / BitStruct(
        'has_focus' / Flag,
        'title_location' / XEnum(BitsInteger(15), enum.ActiveTitleLocation)
    ),
    'product_id' / UUIDAdapter(),
    'sandbox_id' / UUIDAdapter(),
    'aum' / SGString()
)


console_status = XStruct(
    'live_tv_provider' / Int32ub,
    'major_version' / Int32ub,
    'minor_version' / Int32ub,
    'build_number' / Int32ub,
    'locale' / SGString(),
    'active_titles' / PrefixedArray(Int16ub, _active_title)
)


active_surface_change = XStruct(
    'surface_type' / XEnum(Int16ub, enum.ActiveSurfaceType),
    'server_tcp_port' / Int16ub,
    'server_udp_port' / Int16ub,
    'session_id' / UUIDAdapter(),
    'render_width' / Int16ub,
    'render_height' / Int16ub,
    'master_session_key' / Bytes(0x20)
)

message_structs = {
    MessageType.Ack: acknowledge,
    MessageType.Group: Pass,
    MessageType.LocalJoin: local_join,
    MessageType.StopActivity: Pass,
    MessageType.AuxilaryStream: auxiliary_stream,
    MessageType.ActiveSurfaceChange: active_surface_change,
    MessageType.Navigate: Pass,
    MessageType.Json: json,
    MessageType.Tunnel: Pass,
    MessageType.ConsoleStatus: console_status,
    MessageType.TitleTextConfiguration: text_configuration,
    MessageType.TitleTextInput: title_text_input,
    MessageType.TitleTextSelection: title_text_selection,
    MessageType.MirroringRequest: Pass,
    MessageType.TitleLaunch: title_launch,
    MessageType.StartChannelRequest: start_channel_request,
    MessageType.StartChannelResponse: start_channel_response,
    MessageType.StopChannel: stop_channel,
    MessageType.System: Pass,
    MessageType.Disconnect: disconnect,
    MessageType.TitleTouch: touch,
    MessageType.Accelerometer: accelerometer,
    MessageType.Gyrometer: gyrometer,
    MessageType.Inclinometer: inclinometer,
    MessageType.Compass: compass,
    MessageType.Orientation: orientation,
    MessageType.PairedIdentityStateChanged: paired_identity_state_changed,
    MessageType.Unsnap: unsnap,
    MessageType.GameDvrRecord: game_dvr_record,
    MessageType.PowerOff: power_off,
    MessageType.MediaControllerRemoved: media_controller_removed,
    MessageType.MediaCommand: media_command,
    MessageType.MediaCommandResult: media_command_result,
    MessageType.MediaState: media_state,
    MessageType.Gamepad: gamepad,
    MessageType.SystemTextConfiguration: text_configuration,
    MessageType.SystemTextInput: system_text_input,
    MessageType.SystemTouch: touch,
    MessageType.SystemTextAck: system_text_acknowledge,
    MessageType.SystemTextDone: system_text_done
}

struct = XStruct(
    'header' / header,
    'protected_payload' / CryptoTunnel(
        IfThenElse(this.header.flags.is_fragment, fragment,
            XSwitch(
                this.header.flags.msg_type,
                message_structs,
                Pass
            )
        )
    )
)



# flake8: noqa
"""
Construct containers for simple-message header and payloads
"""
from construct import *
from xbox.sg import enum
from xbox.sg.enum import PacketType
from xbox.sg.utils.struct import XStruct
from xbox.sg.utils.adapters import CryptoTunnel, XSwitch, XEnum, CertificateAdapter, UUIDAdapter, SGString, FieldIn


pkt_types = [
    PacketType.PowerOnRequest,
    PacketType.DiscoveryRequest, PacketType.DiscoveryResponse,
    PacketType.ConnectRequest, PacketType.ConnectResponse
]


header = XStruct(
    'pkt_type' / XEnum(Int16ub, PacketType),
    'unprotected_payload_length' / Default(Int16ub, 0),
    'protected_payload_length' / If(
        FieldIn('pkt_type', [PacketType.ConnectRequest, PacketType.ConnectResponse]),
        Default(Int16ub, 0)
    ),
    'version' / Default(Int16ub, 2)
)


power_on_request = XStruct(
    'liveid' / SGString()
)


discovery_request = XStruct(
    'flags' / Int32ub,
    'client_type' / XEnum(Int16ub, enum.ClientType),
    'minimum_version' / Int16ub,
    'maximum_version' / Int16ub
)


discovery_response = XStruct(
    'flags' / XEnum(Int32ub, enum.PrimaryDeviceFlag),
    'type' / XEnum(Int16ub, enum.ClientType),
    'name' / SGString(),
    'uuid' / UUIDAdapter('utf8'),
    'last_error' / Int32ub,
    'cert' / CertificateAdapter()
)


connect_request_unprotected = XStruct(
    'sg_uuid' / UUIDAdapter(),
    'public_key_type' / XEnum(Int16ub, enum.PublicKeyType),
    'public_key' / XSwitch(this.public_key_type, {
        enum.PublicKeyType.EC_DH_P256: Bytes(0x40),
        enum.PublicKeyType.EC_DH_P384: Bytes(0x60),
        enum.PublicKeyType.EC_DH_P521: Bytes(0x84)
    }),
    'iv' / Bytes(0x10)
)


connect_request_protected = XStruct(
    'userhash' / SGString(),
    'jwt' / SGString(),
    'connect_request_num' / Int32ub,
    'connect_request_group_start' / Int32ub,
    'connect_request_group_end' / Int32ub
)


connect_response_unprotected = XStruct(
    'iv' / Bytes(0x10)
)


connect_response_protected = XStruct(
    'connect_result' / XEnum(Int16ub, enum.ConnectionResult),
    'pairing_state' / XEnum(Int16ub, enum.PairedIdentityState),
    'participant_id' / Int32ub
)


struct = XStruct(
    'header' / header,
    'unprotected_payload' / XSwitch(
        this.header.pkt_type, {
            PacketType.PowerOnRequest: power_on_request,
            PacketType.DiscoveryRequest: discovery_request,
            PacketType.DiscoveryResponse: discovery_response,
            PacketType.ConnectRequest: connect_request_unprotected,
            PacketType.ConnectResponse: connect_response_unprotected
        }
    ),
    'protected_payload' / CryptoTunnel(
        XSwitch(
            this.header.pkt_type, {
                PacketType.ConnectRequest: connect_request_protected,
                PacketType.ConnectResponse: connect_response_protected
            },
            Pass
        )
    )
)



from http import HTTPStatus


def test_auth_overview(rest_client):
    resp = rest_client.test_client().get('/auth')
    json = resp.json

    assert resp.status_code == HTTPStatus.OK
    assert json is not None

    assert 'success' in json
    assert 'authenticated' in json
    assert 'tokens' in json
    assert 'userinfo' in json

    assert 'access_token' in json['tokens']
    assert 'refresh_token' in json['tokens']
    assert 'user_token' in json['tokens']
    assert 'xsts_token' in json['tokens']

    assert json['success'] is True


def test_auth_login_get(rest_client):
    resp = rest_client.test_client().get('/auth/login')

    assert resp.status_code == HTTPStatus.OK
    assert resp.json is None
    assert b'Authenticate with Windows Live' in resp.data


def test_auth_login_post_no_params(rest_client):
    # No post params
    resp = rest_client.test_client().post('/auth/login')

    assert resp.status_code == HTTPStatus.BAD_REQUEST
    assert resp.json['success'] is False
    assert resp.json['message'] == 'No email or password parameter provided'


def test_auth_login_post_invalid_credentials(rest_client):
    resp = rest_client.test_client().post('/auth/login',
                                          data={'email': 'foo@bar.com', 'password': '123'})

    assert resp.status_code == HTTPStatus.INTERNAL_SERVER_ERROR
    assert resp.json['success'] is False
    assert 'Login failed' in resp.json['message']
    assert resp.json['two_factor_required'] is False


def test_auth_login_post_webview_invalid_credentials(rest_client):
    resp = rest_client.test_client().post('/auth/login',
                                          data={'webview': True, 'email': 'foo@bar.com',
                                                'password': '123'})

    assert resp.status_code == HTTPStatus.OK
    assert resp.json is None
    assert b'Login failed' in resp.data


def test_auth_logout_get_not_logged_in(rest_client):
    resp = rest_client.test_client().get('/auth/logout')

    assert resp.status_code == HTTPStatus.OK
    assert resp.json is None
    assert b'currently not logged in' in resp.data


def test_auth_logout_post(rest_client):
    resp = rest_client.test_client().post('/auth/logout')

    assert resp.status_code == HTTPStatus.OK
    assert resp.json['success'] is True
    assert resp.json['message'] == 'Logout succeeded'


def test_auth_logout_post_webview(rest_client):
    resp = rest_client.test_client().post('/auth/logout', data={'webview': True})

    assert resp.status_code == HTTPStatus.OK
    assert resp.json is None
    assert b'Logout succeeded' in resp.data


def test_auth_url(rest_client):
    resp = rest_client.test_client().get('/auth/url')

    assert resp.status_code == HTTPStatus.OK
    assert resp.json['success'] is True
    assert resp.json['authorization_url'] is not None
    assert resp.json['authorization_url'].startswith('https://login.live.com/oauth20_authorize.srf')


def test_auth_oauth_get(rest_client):
    resp = rest_client.test_client().get('/auth/oauth')

    assert resp.status_code == HTTPStatus.OK
    assert resp.json is None
    assert b'Login via OAUTH' in resp.data


def test_auth_oauth_post_no_params(rest_client):
    resp = rest_client.test_client().post('/auth/oauth')

    assert resp.status_code == HTTPStatus.BAD_REQUEST
    assert resp.json['success'] is False
    assert resp.json['message'] == 'Please provide redirect_url'


def test_auth_oauth_post_no_params_webview(rest_client):
    resp = rest_client.test_client().post('/auth/oauth', data={'webview': True})

    assert resp.status_code == HTTPStatus.BAD_REQUEST
    assert resp.json['success'] is False
    assert resp.json['message'] == 'Please provide redirect_url'


def test_auth_oauth_post_invalid_params(rest_client):
    resp = rest_client.test_client().post('/auth/oauth', data={'redirect_uri': 'hxxxp:/invalid'})

    assert resp.status_code == HTTPStatus.INTERNAL_SERVER_ERROR
    assert resp.json['success'] is False
    assert resp.json['message'].startswith('Login failed')


def test_auth_oauth_post_invalid_params_webview(rest_client):
    resp = rest_client.test_client().post('/auth/oauth', data={'webview': True,
                                          'redirect_uri': 'hxxxp:/invalid'})

    assert resp.status_code == HTTPStatus.OK
    assert b'Login failed' in resp.data



"""
Adapters and other Construct utility classes
"""
import json
import construct
from io import BytesIO
from enum import Enum
from uuid import UUID
from cryptography import x509
from cryptography.x509.oid import NameOID
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.serialization import Encoding

from xbox.sg.enum import PacketType


class CryptoTunnel(construct.Subconstruct):
    """
    Adapter/Tunnel for inline decryption of protected payloads.

    Depending on the packet-type, acquiring the `Initialization Vector` and
    decrypting the `protected payload` happens differently:
    * ConnectResponse: The `IV` is delivered in the `unprotected payload`
    section of the packet and can be used directly
    * Messages: The `IV` is generated by encrypting the first 16 bytes of
    the packet header using the IV key.

    Inline encryption using this Tunnel is not (yet) possible due to
    limitations in Construct. As explained above, the `IV` for Messages is
    generated by encrypting the first 16 bytes of the header. However, there's
    currently no method of determining the length of the payload without
    building it first. Therefor, the `IV` wouldn't be correct.
    """
    def _parse(self, stream, context, path):
        return self.subcon._parse(self.decrypt(stream, context), context, path)

    @staticmethod
    def decrypt(stream, context):
        # To simplify message definition, we allow Switch to be a subcon
        # of CryptoAdapter
        # However, as a side effect, this function will also be called when the
        # stream is EOF (has no protected payload)
        # To compensate for this, we check if there's at least a hash + 1 block
        # left in the stream
        pos = stream.tell()
        if len(stream.read(48)) != 48:
            return None

        crypto = context.get('_crypto', None) or context._.get('_crypto', None)

        if not crypto:
            raise ValueError("Crypto instance not passed in context")

        stream.seek(0)
        buf = stream.read()

        if not crypto.verify(buf[:-32], buf[-32:]):
            raise ValueError("Checksum doesn't match")

        connect_types = [PacketType.ConnectRequest, PacketType.ConnectResponse]
        if context.header.pkt_type in connect_types:
            iv = context.unprotected_payload.iv
        elif context.header.pkt_type == PacketType.Message:
            iv = crypto.generate_iv(buf[:16])
        else:
            raise ValueError("Incompatible packet type")

        buf = buf[pos:]
        decrypted = crypto.decrypt(iv, buf)
        decrypted = decrypted[:context.header.protected_payload_length]

        return BytesIO(decrypted)

    def _emitparse(self, code):
        # Hack
        code.append('from xbox.sg.utils.adapters import CryptoTunnel')
        subcode = self.subcon._emitparse(code)
        return subcode.replace('(io', '(CryptoTunnel.decrypt(io, this)')


class JsonAdapter(construct.Adapter):
    """
    Construct-Adapter for JSON field.

    Parses and dumps JSON.
    """
    def _encode(self, obj, context, path):
        if not isinstance(obj, dict):
            raise TypeError('Object not of type dict')
        return json.dumps(obj, separators=(',', ':'), sort_keys=True)

    def _decode(self, obj, context, path):
        return json.loads(obj)

    def _emitparse(self, code):
        code.append('import json')
        return 'json.loads({})'.format(self.subcon._emitparse(code))


class UUIDAdapter(construct.Adapter):
    def __init__(self, encoding=None):
        """
        Construct-Adapter for UUID field.

        Parses either `utf8` encoded or raw byte strings into :class:`UUID`
        instances.

        Args:
            encoding (str): The encoding to use.
        """
        if encoding:
            super(self.__class__, self).__init__(SGString(encoding))
        else:
            super(self.__class__, self).__init__(construct.Bytes(0x10))

    def _encode(self, obj, context, path):
        if not isinstance(obj, UUID):
            raise TypeError('Object not of type UUID')
        if isinstance(self.subcon, construct.Bytes):
            return obj.bytes
        else:
            return str(obj).upper()

    def _decode(self, obj, context, path):
        if isinstance(self.subcon, construct.Bytes):
            return UUID(bytes=obj)
        else:
            return UUID(obj)

    def _emitparse(self, code):
        code.append('from uuid import UUID')
        subcon_code = self.subcon._emitparse(code)
        if isinstance(self.subcon, construct.Bytes):
            return 'UUID(bytes={})'.format(subcon_code)
        return 'UUID({})'.format(subcon_code)


class CertificateAdapter(construct.Adapter):
    def __init__(self):
        """
        Construct-Adapter for Certificate field.

        Parses and dumps the DER certificate as used in the discovery response
        messages.
        """
        super(self.__class__, self).__init__(PrefixedBytes(construct.Int16ub))

    def _encode(self, obj, context, path):
        if not isinstance(obj, CertificateInfo):
            raise TypeError('Object not of type CertificateInfo')
        return obj.dump()

    def _decode(self, obj, context, path):
        return CertificateInfo(obj)

    def _emitparse(self, code):
        code.append('from xbox.sg.utils.adapters import CertificateInfo')
        return 'CertificateInfo({})'.format(self.subcon._emitparse(code))


class CertificateInfo(object):
    def __init__(self, raw_cert):
        """
        Helper class for parsing a x509 certificate.

        Extracts `common_name` and `public_key` from the certificate.

        Args:
            raw_cert (bytes): The DER certificate to parse.
        """
        self.cert = x509.load_der_x509_certificate(raw_cert, default_backend())
        self.liveid = self.cert.subject.get_attributes_for_oid(
            NameOID.COMMON_NAME)[0].value
        self.pubkey = self.cert.public_key()

    def dump(self, encoding=Encoding.DER):
        return self.cert.public_bytes(encoding)

    def __repr__(self):
        return '<%s: liveid=%s, pubkey=%s>' % (
            self.__class__.__name__, self.liveid, self.pubkey
        )

    def __eq__(self, other):
        return self.cert == other.cert


class XSwitch(construct.Switch):
    def _emitparse(self, code):
        if not any([isinstance(k, Enum) for k in self.cases.keys()]):
            return super(XSwitch, self)._emitparse(code)

        fname = "factory_%s" % code.allocateId()
        code.append("%s = {%s}" % (fname, ", ".join("%r : lambda io,this: %s" % (key.value, sc._compileparse(code))
                                                    for key, sc in self.cases.items()), ))

        defaultfname = "compiled_%s" % code.allocateId()
        code.append("%s = lambda io,this: %s" % (defaultfname, self.default._compileparse(code), ))
        return "%s.get(%s.value, %s)(io, this)" % (fname, self.keyfunc, defaultfname)


class XEnum(construct.Adapter):
    def __init__(self, subcon, enum=None):
        """
        Construct-Adapter for Enum field.

        Parses numeric fields into `XEnumInt`'s, which display the Enum name and value.

        Args:
            subcon (Construct): The subcon to adapt.
            enum (Enum): The enum to parse into.
        """
        super(XEnum, self).__init__(subcon)
        self.enum = enum

    def _encode(self, obj, context, path):
        if isinstance(obj, int):
            return obj
        return obj.value

    def _decode(self, obj, context, path):
        if not self.enum:
            return obj
        return self.enum(obj)

    def _emitparse(self, code):
        if not self.enum:
            return self.subcon._emitparse(code)

        code.append('from {} import {}'.format(self.enum.__module__, self.enum.__name__))
        return '{}({})'.format(self.enum.__name__, self.subcon._emitparse(code))


class XInject(construct.Construct):
    def __init__(self, code):
        super(XInject, self).__init__()
        self.flagbuildnone = True
        self.name = '__inject'
        self.code = code

    def _parse(self, stream, context, path):
        pass

    def _build(self, obj, stream, context, path):
        pass

    def _sizeof(self, context, path):
        return 0

    def _emitparse(self, code):
        code.append(self.code)


class TerminatedField(construct.Subconstruct):
    def __init__(self, subcon, length=1, pattern=b'\x00'):
        """
        A custom :class:`Subconstruct` that adds a termination character at
        the end of the child struct.

        Args:
            subcon (Construct): The subcon to add the terminated character to.
            length (int): The amount of termination characters to add.
            pattern (bytes): The termination pattern to use.
        """
        super(self.__class__, self).__init__(subcon)
        self.padding = construct.Padding(length, pattern)

    def _parse(self, stream, context, path):
        obj = self.subcon._parse(stream, context, path)
        self.padding._parse(stream, context, path)

        return obj

    def _build(self, obj, stream, context, path):
        subobj = self.subcon._build(obj, stream, context, path)
        self.padding._build(obj, stream, context, path)

        return subobj

    def _sizeof(self, context, path):
        return self.subcon._sizeof(context, path) + 1

    def _emitparse(self, code):
        return '({}, {})[0]'.format(
            self.subcon._emitparse(code), self.padding._emitparse(code)
        )


def SGString(encoding='utf8'):
    """
    Defines a null terminated `PascalString`.

    The Smartglass protocol seems to always add a termination character
    behind a length prefixed string.
    This utility function combines a `PascalString` with a `TerminatedField`.

    Args:
        encoding (str): The string encoding to use for the `PascalString`.

    Returns:
        SGString: A null byte terminated `PascalString`.
    """
    return TerminatedField(construct.PascalString(construct.Int16ub, encoding))


def PrefixedBytes(lengthfield):
    """
    Defines a length prefixed bytearray.

    Args:
        lengthfield (:class:`Subconstruct`): The length subcon.

    Returns:
        PrefixedBytes: A length prefixed bytesarray
    """
    return construct.Prefixed(lengthfield, construct.GreedyBytes)


class FieldIn(object):
    """
    Helper class for creating an `in` conditional.

    Operates like `field in options`.

    Args:
        field: The struct field to use.
        options (list): A list with options to execute the `in` conditional on.

    Returns:
        bool: Whether or not the field value was in the options.
    """
    def __init__(self, field, options):
        self.field = field
        self.options = options

    def __call__(self, context):
        return context.get(self.field, None) in self.options

    def __repr__(self):
        if any([isinstance(e, Enum) for e in self.options]):
            return 'this.{}.value in {}'.format(self.field, [e.value for e in self.options])
        return 'this.{} in {}'.format(self.field, self.options)



"""
JSON models for deserializing Stump messages
"""
from marshmallow_objects import Model, NestedModel, fields
from xbox.stump.enum import Message


class StumpJsonError(Exception):
    pass


# Root-level containers
class StumpResponse(Model):
    msgid = fields.Str()
    response = fields.Str()


class StumpError(Model):
    msgid = fields.Str()
    error = fields.Str()


class StumpNotification(Model):
    notification = fields.Str()


# Nested models
class _FoundChannel(Model):
    channelNumber = fields.Int()
    displayName = fields.Str()
    channelId = fields.Str()


class _LineupProvider(Model):
    foundChannels = NestedModel(_FoundChannel, many=True, required=True)
    cqsChannels = fields.List(fields.Str(), required=True)
    headendId = fields.UUID()


class _EnsureStreamingStarted(Model):
    currentChannelId = fields.Str()
    source = fields.Str()
    streamingPort = fields.Int()
    tunerChannelType = fields.Str()
    userCanViewChannel = fields.Str()


class _TunerLineups(Model):
    providers = NestedModel(_LineupProvider, many=True)


class _RecentChannel(Model):
    channelNum = fields.Str()  # Can be "NumberUnused" instead of int
    providerId = fields.UUID()
    channelId = fields.Str()


class _PauseBufferInfo(Model):
    Enabled = fields.Bool()
    IsDvr = fields.Bool()
    MaxBufferSize = fields.Int()
    BufferCurrent = fields.Int()
    BufferStart = fields.Int()
    BufferEnd = fields.Int()
    CurrentTime = fields.Int()
    Epoch = fields.Int()


class _LiveTvInfo(Model):
    streamingPort = fields.Int()
    inHdmiMode = fields.Bool()
    tunerChannelType = fields.Str()
    currentTunerChannelId = fields.Str()
    currentHdmiChannelId = fields.Str()
    pauseBufferInfo = NestedModel(_PauseBufferInfo)


class _HeadendProvider(Model):
    providerName = fields.Str()
    filterPreference = fields.Str()
    headendId = fields.UUID()
    source = fields.Str()
    titleId = fields.Str()
    canStream = fields.Bool()


class _HeadendInfo(Model):
    providerName = fields.Str()
    headendId = fields.UUID()
    blockExplicitContentPerShow = fields.Bool()
    dvrEnabled = fields.Bool()
    headendLocale = fields.Str()
    streamingPort = fields.Int()
    preferredProvider = fields.Str()
    providers = NestedModel(_HeadendProvider, many=True)


class _DeviceConfiguration(Model):
    device_id = fields.Str()
    device_type = fields.Str()
    device_brand = fields.Str()
    device_model = fields.Str()
    device_name = fields.Str()
    buttons = fields.Dict()


class _AppChannel(Model):
    name = fields.Str()
    id = fields.Str()


class _AppProvider(Model):
    id = fields.Str()
    providerName = fields.Str()
    titleId = fields.Str()
    primaryColor = fields.Str()
    secondaryColor = fields.Str()
    providerImageUrl = fields.Str()
    channels = NestedModel(_AppChannel, many=True)


# Stump responses
class AppChannelLineups(StumpResponse):
    params = NestedModel(_AppProvider, many=True)


class EnsureStreamingStarted(StumpResponse):
    params = NestedModel(_EnsureStreamingStarted)


class TunerLineups(StumpResponse):
    params = NestedModel(_TunerLineups)


class SendKey(StumpResponse):
    params = fields.Bool()


class RecentChannels(StumpResponse):
    params = NestedModel(_RecentChannel, many=True)


class Configuration(StumpResponse):
    params = NestedModel(_DeviceConfiguration, many=True)


class LiveTvInfo(StumpResponse):
    params = NestedModel(_LiveTvInfo)


class HeadendInfo(StumpResponse):
    params = NestedModel(_HeadendInfo)


response_map = {
    Message.CONFIGURATION: Configuration,
    Message.ENSURE_STREAMING_STARTED: EnsureStreamingStarted,
    Message.SEND_KEY: SendKey,
    Message.RECENT_CHANNELS: RecentChannels,
    Message.SET_CHANNEL: None,
    Message.APPCHANNEL_PROGRAM_DATA: None,
    Message.APPCHANNEL_DATA: None,
    Message.APPCHANNEL_LINEUPS: AppChannelLineups,
    Message.TUNER_LINEUPS: TunerLineups,
    Message.PROGRAMM_INFO: None,
    Message.LIVETV_INFO: LiveTvInfo,
    Message.HEADEND_INFO: HeadendInfo,
    Message.ERROR: None
}


def deserialize_stump_message(data):
    """
    Helper for deserializing JSON stump messages

    Args:
        data (dict): Stump message

    Returns:
        Model: Parsed JSON object
    """
    response = data.get('response')
    notification = data.get('notification')
    error = data.get('error')

    if response:
        response = Message(response)
        model = response_map.get(response)
        if not issubclass(model, StumpResponse):
            raise StumpJsonError('Model not of subclass StumpResponse')

        return model.load(data)
    elif notification:
        return StumpNotification.load(data)
    elif error:
        return StumpError.load(data)
